/*
 * ksocket.c
 *
 * This file reimplements the socket layer in a kernel-friendly way, allowing
 * drivers and modules to use the network stack.
 *
 * It is based heavily upon net/socket.c.
 *
 * Version:	0.10
 * Date:	Sep 27 2004
 * Author:	Stuart MacDonald <stuartm@connecttech.com>
 *
 */

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>

#include <linux/net.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <asm/uaccess.h>
#include <net/sock.h>
#include <linux/init.h>

#include "bhn_compat.h"

#define MAX_SOCK_ADDR	128

static int __init ksocket_init(void)
{
	return 0;
}

static void __exit ksocket_fini(void)
{
	return;
}

module_init(ksocket_init);
module_exit(ksocket_fini);
#ifndef NO_SOMAXCONN
MODULE_DESCRIPTION("Kernel-friendly socket layer");
#else
MODULE_DESCRIPTION("Kernel-friendly socket layer; ignores sysctl_somaxconn");
#endif
MODULE_AUTHOR("Stuart MacDonald <stuartm@connecttech.com>");
MODULE_LICENSE("GPL");

int ksys_socket(int family, int type, int protocol, struct socket **sock)
{
	return sock_create(family, type, protocol, sock);
}

int ksys_socketpair(int family, int type, int protocol,
						struct socket *sockvec[2])
{
	struct socket *sock1;
	struct socket *sock2;
	int err;

	/*
	 * Obtain the first socket and check if the underlying protocol
	 * supports the socketpair call.
	 */

	err = sock_create(family, type, protocol, &sock1);
	if (err < 0)
		goto out;

	err = sock_create(family, type, protocol, &sock2);
	if (err < 0)
		goto out_release_1;

	err = sock1->ops->socketpair(sock1, sock2);
	if (err < 0)
		goto out_release_both;

	sockvec[0] = sock1;
	sockvec[1] = sock2;
	return 0;

out_release_both:
        sock_release(sock2);
out_release_1:
        sock_release(sock1);
out:
	return err;
}

int ksys_bind(struct socket *sock, struct sockaddr *addr, int addrlen)
{
	return sock->ops->bind(sock, addr, addrlen);
}

#ifndef NO_SOMAXCONN
extern int sysctl_somaxconn;
#endif

int ksys_listen(struct socket *sock, int backlog)
{
	int err;

#ifndef NO_SOMAXCONN
	if ((unsigned) backlog > sysctl_somaxconn)
		backlog = sysctl_somaxconn;
#endif
	err = sock->ops->listen(sock, backlog);

	return err;
}

int ksys_accept(struct socket *sock, struct sockaddr *peer_sockaddr,
						int *peer_addrlen, int flags)
{
	struct socket *newsock;
	int err, len;
	char address[MAX_SOCK_ADDR];

	err = -EMFILE;
	if (!(newsock = sock_alloc()))
		goto out;

	newsock->type = sock->type;
	newsock->ops = sock->ops;

	err = sock->ops->accept(sock, newsock, flags);
	if (err < 0)
		goto out_release;

	if (peer_sockaddr) {
		if (newsock->ops->getname(newsock, (struct sockaddr *)address,
								&len, 2) < 0) {
			err = -ECONNABORTED;
			goto out_release;
		}
		memcpy(peer_sockaddr, address, *peer_addrlen);
	}

out:
	return err;
out_release:
	sock_release(newsock);
	goto out;
}

int ksys_connect(struct socket *sock, struct sockaddr *addr, int addrlen,
								int flags)
{
	return sock->ops->connect(sock, addr, addrlen, flags);
}

int ksys_getsockname(struct socket *sock, struct sockaddr *sockaddr,
							int *sockaddr_len)
{
	char address[MAX_SOCK_ADDR];
	int len, err;

	err = sock->ops->getname(sock, (struct sockaddr *)address, &len, 0);
	if (!err)
		memcpy(sockaddr, address, *sockaddr_len);
	return err;
}

int ksys_getpeername(struct socket *sock, struct sockaddr *sockaddr,
							int *sockaddr_len)
{
	char address[MAX_SOCK_ADDR];
	int len, err;

	err = sock->ops->getname(sock, (struct sockaddr *)address, &len, 1);
	if (!err)
		memcpy(sockaddr, address, *sockaddr_len);
	return err;
}

int ksys_sendto(struct socket *sock, void *buff, size_t len, unsigned flags,
			   		struct sockaddr *addr, int addr_len)
{
	int err;
	struct msghdr msg;
	struct iovec iov;
	mm_segment_t old_fs;

	iov.iov_base = buff;
	iov.iov_len = len;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags = flags;
	if (addr) {
		msg.msg_name = addr;
		msg.msg_namelen = addr_len;
	} else {
		msg.msg_name = NULL;
		msg.msg_namelen = 0;
	}

	old_fs = get_fs();
	set_fs(KERNEL_DS);

	err = sock_sendmsg(sock, &msg, len);

	set_fs(old_fs);

	return err;
}

int ksys_send(struct socket *sock, void *buff, size_t len, unsigned flags)
{
	return ksys_sendto(sock, buff, len, flags, NULL, 0);
}

int ksys_recvfrom(struct socket *sock, void *ubuf, size_t size, unsigned flags,
			     struct sockaddr *addr, int *addr_len)
{
	struct iovec iov;
	struct msghdr msg;
	char address[MAX_SOCK_ADDR];
	int err;
	mm_segment_t old_fs;

	iov.iov_base = ubuf;
	iov.iov_len = size;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_name = address;
	msg.msg_namelen = MAX_SOCK_ADDR;

	old_fs = get_fs();
	set_fs(KERNEL_DS);

	err = sock_recvmsg(sock, &msg, size, flags);

	set_fs(old_fs);

	if (err >= 0 && addr != NULL)
		memcpy(addr, address, *addr_len);

	return err;
}

int ksys_recv(struct socket *sock, void *ubuf, size_t size, unsigned flags)
{
	return ksys_recvfrom(sock, ubuf, size, flags, NULL, NULL);
}

int ksys_setsockopt(struct socket *sock, int level, int optname, char *optval,
								int optlen)
{
	mm_segment_t old_fs;
	int err;

	if (optlen < 0)
		return -EINVAL;

	old_fs = get_fs();
	set_fs(KERNEL_DS);

	if (level == SOL_SOCKET)
		err = sock_setsockopt(sock, level, optname, optval, optlen);
	else
		err = sock->ops->setsockopt(sock, level, optname, optval,
									optlen);

	set_fs(old_fs);

	return err;
}

int ksys_getsockopt(struct socket *sock, int level, int optname, char *optval,
								int *optlen)
{
	int err;
	mm_segment_t old_fs;

	old_fs = get_fs();
	set_fs(KERNEL_DS);

	if (level == SOL_SOCKET)
		err = sock_getsockopt(sock, level, optname, optval, optlen);
	else
		err = sock->ops->getsockopt(sock, level, optname, optval,
									optlen);

	set_fs(old_fs);

	return err;
}

int ksys_shutdown(struct socket *sock, int how)
{
	return sock->ops->shutdown(sock, how);
}

int ksys_sendmsg(struct socket *sock, struct msghdr *msg, unsigned int flags)
{
	int err, total_len;
	mm_segment_t old_fs;
	int i;

	/* do not move before msg is valid */
	err = -EMSGSIZE;
	if (msg->msg_iovlen > UIO_MAXIOV)
		goto out;

	err = -ENOBUFS;
	if (msg->msg_controllen > INT_MAX)
		goto out;

	msg->msg_flags = flags;

	for (total_len = 0, i = 0; i < msg->msg_iovlen; i++)
		total_len += msg->msg_iov[i].iov_len;

	old_fs = get_fs();
	set_fs(KERNEL_DS);

	err = sock_sendmsg(sock, msg, total_len);

	set_fs(old_fs);

out:
	return err;
}

int ksys_recvmsg(struct socket *sock, struct msghdr *msg, unsigned int flags)
{
	int err, total_len;
	int i;
	mm_segment_t old_fs;

	err = -EMSGSIZE;
	if (msg->msg_iovlen > UIO_MAXIOV)
		goto out;

	msg->msg_flags = 0;

	for (total_len = 0, i = 0; i < msg->msg_iovlen; i++)
		total_len += msg->msg_iov[i].iov_len;

	old_fs = get_fs();
	set_fs(KERNEL_DS);

	err = sock_recvmsg(sock, msg, total_len, flags);

	set_fs(old_fs);

out:
	return err;
}

EXPORT_SYMBOL(ksys_socket);
EXPORT_SYMBOL(ksys_socketpair);
EXPORT_SYMBOL(ksys_bind);
EXPORT_SYMBOL(ksys_listen);
EXPORT_SYMBOL(ksys_accept);
EXPORT_SYMBOL(ksys_connect);
EXPORT_SYMBOL(ksys_getsockname);
EXPORT_SYMBOL(ksys_getpeername);
EXPORT_SYMBOL(ksys_sendto);
EXPORT_SYMBOL(ksys_send);
EXPORT_SYMBOL(ksys_recvfrom);
EXPORT_SYMBOL(ksys_recv);
EXPORT_SYMBOL(ksys_setsockopt);
EXPORT_SYMBOL(ksys_getsockopt);
EXPORT_SYMBOL(ksys_shutdown);
EXPORT_SYMBOL(ksys_sendmsg);
EXPORT_SYMBOL(ksys_recvmsg);
