/*
 *
 *  The Blue Heat/Net driver's protocol processing
 *
 *  Copyright (C) 2004 Connect Tech Inc.
 *
 *  Oct 14 2004 Stuart MacDonald <stuartm@connecttech.com>
 *	First draft.
 *
 */

#include <linux/version.h>
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <net/sock.h>
#include <linux/tty.h>
#include <linux/circ_buf.h>
#include <net/tcp.h>
#include <linux/completion.h>

#include "ksocket.h"

#include <BHN_flash.h>
#include <CDS.h>
#include <BHN_protocol.h>
#include "bhn.h"
#define NEED_WAIT_FOR_COMPLETION_INTERRUPTIBLE_TIMEOUT
#include "bhn_compat.h"

/* should be moved to BHN_protocol.h at some point */
#define CMD_LOCAL_NETDOWN	213

extern struct tty_driver bhn_driver;

static int baud_table[] = {
	0, 50, 75, 110, 134, 150, 200, 300,
	600, 1200, 1800, 2400, 4800, 9600, 19200,
	38400, 57600, 115200, 230400, 460800, 0 };

static int n_baud_table = sizeof(baud_table) / sizeof(int);

static unsigned int tty_get_baud_idx(int baud)
{
	unsigned int cflag = -1;
	int i;

	for (i = 0; i < n_baud_table; i++)
		if (baud == baud_table[i])
			break;
	if (i != n_baud_table) {
		cflag = 0;
		if (i > 15) {
			cflag |= CBAUDEX;
			i -= 15;
		}
		cflag |= i;
	}
	return cflag;
}

static int is_complete(struct completion *x)
{
	unsigned long flags;
	int ret;

	spin_lock_irqsave(&x->wait.lock, flags);
	ret = (x->done != 0);
	spin_unlock_irqrestore(&x->wait.lock, flags);

	return ret;
}

/*
 * Protocol processing
 */

int prot_tcp_start(int minor,
		void (*fn)(int minor, struct bhn_event_occurred * events))
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_unit *unit = map->unit;
	struct bhn_state *state;
	int timeout;
	int pid;
	int err;

	down(&map->auto_reconn_lock);
	pid = map->auto_reconn_pid;
	up(&map->auto_reconn_lock);

	if (pid)
		state = map->state;
	else {
		if (map->state) {
			printk("%s: Error: State on TCP start\n", bhn_drvname);
			prot_tcp_stop(minor);
		}

		state = kmalloc(sizeof(*state), GFP_KERNEL);
		if (!state) {
			err = -ENOMEM;
			goto out;
		}
		memset(state, 0, sizeof(*state));
		state->cmd.addr = unit->addr;
		state->cmd.addr.sin_port = htons(CMDPORT(map));
		init_completion(&state->cmd.read_comp);
		init_completion(&state->cmd.write_comp);
		state->data.addr = unit->addr;
		state->data.addr.sin_port = htons(DATAPORT(map));
		init_completion(&state->data.read_comp);
		init_completion(&state->data.write_comp);
		init_waitqueue_head(&state->data.wq_write);
		INIT_LIST_HEAD(&state->cmdlist);
		spin_lock_init(&state->listlock);
		init_waitqueue_head(&state->wq_throttle);
		init_waitqueue_head(&state->wq_purge);
		state->event_cb = fn;
		map->state = state;
	}

	if (map->flags & MAP_AUTO_RECONNECT) {
		timeout = map->cmd_timeout / HZ;
		if (!timeout)
			timeout = 1;
	} else
		timeout = 0;

	down(&map->cr_lock);
	down(&map->cw_lock);
	err = prot_tcp_connect(&state->cmd.sock, (SOCKADDR)&state->cmd.addr,
								timeout);
	up(&map->cw_lock);
	up(&map->cr_lock);
	if (err < 0) {
		down(&map->auto_reconn_lock);
		goto free_out;
	}

	err = kernel_thread(prot_tcp_cmdread, (void *)minor, 0);
	if (err < 0) {
		down(&map->auto_reconn_lock);
		goto connect_out;
	}
	/*
	 * If we don't wait for startup, it's possible that the kill_proc()
	 * can happen before the process has started, and that SIGKILL gets
	 * lost resulting in hung user space apps.
	 */
	wait_for_completion(&state->cmd.read_comp);
	down(&map->auto_reconn_lock);
	state->cmd.read_pid = err;
	map->flags |= MAP_NETWORK_ALIVE;
	up(&map->auto_reconn_lock);
	err = 0;
	goto out;

connect_out:
	down(&map->cr_lock);
	down(&map->cw_lock);
	if (state->cmd.sock) {
		sock_release(state->cmd.sock);
		state->cmd.sock = NULL;
	}
	up(&map->cw_lock);
	up(&map->cr_lock);
free_out:
	if (!map->auto_reconn_pid) {
		kfree(state);
		map->state = NULL;
	}
	up(&map->auto_reconn_lock);

out:
	return err;
}

int prot_tcp_stop(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;

	/*
	 * This can happen legitimately from a failed open, the TTY layer
	 * calls close() anyway.
	 */
	if (!state)
		goto out;

	down(&map->auto_reconn_lock);
	map->flags &= ~MAP_NETWORK_ALIVE;
	up(&map->auto_reconn_lock);

	if (state->cmd.read_pid) {
		kill_proc(state->cmd.read_pid, SIGKILL, 1);
		wait_for_completion(&state->cmd.read_comp);
	}
	down(&map->cr_lock);
	down(&map->cw_lock);
	if (state->cmd.sock) {
		sock_release(state->cmd.sock);
		state->cmd.sock = NULL;
	}
	up(&map->cw_lock);
	up(&map->cr_lock);

	down(&map->auto_reconn_lock);
	if (!map->auto_reconn_pid) {
		kfree(state);
		map->state = NULL;
	}
	up(&map->auto_reconn_lock);

out:
	return 0;
}

int prot_tcp_connect(struct socket **sock, struct sockaddr *addr, int timeout)
{
	int err;
	int on = 1;

	err = ksys_socket(PF_INET, SOCK_STREAM, 0, sock);
	if (err < 0)
		goto out;

	err = ksys_setsockopt(*sock, SOL_TCP, TCP_NODELAY, (char *) &on,
								sizeof(on));
	if (err < 0)
		goto release_out;

	if (timeout) {
		int mintime;

		err = ksys_setsockopt(*sock, SOL_SOCKET, SO_KEEPALIVE,
						(char *) &on, sizeof(on));
		if (err < 0)
			goto release_out;
		mintime = MIN(timeout, MAX_TCP_KEEPIDLE);
		err = ksys_setsockopt(*sock, SOL_TCP, TCP_KEEPIDLE,
					(char *) &mintime, sizeof(mintime));
		if (err < 0)
			goto release_out;
		mintime = MIN(timeout, MAX_TCP_KEEPINTVL);
		err = ksys_setsockopt(*sock, SOL_TCP, TCP_KEEPINTVL,
					(char *) &mintime, sizeof(mintime));
		if (err < 0)
			goto release_out;
		err = ksys_setsockopt(*sock, SOL_TCP, TCP_KEEPCNT,
						(char *) &on, sizeof(on));
		if (err < 0)
			goto release_out;
	}

	err = ksys_connect(*sock, addr, sizeof(*addr), 0);
	if (err < 0)
		goto release_out;

	goto out;

release_out:
	if (*sock) {
		sock_release(*sock);
		*sock = NULL;
	}
out:
	return err;
}

static void prot_network_not_alive(int minor) {
	struct bhn_map *map = &bhn_table[minor];
	int err;

	down(&map->auto_reconn_lock);
	map->flags &= ~MAP_NETWORK_ALIVE;
	if (!((map->flags & MAP_OPENING) || (map->flags & MAP_CLOSING)) &&
					(map->flags & MAP_AUTO_RECONNECT)) {
		if (!map->auto_reconn_pid) {
			err = kernel_thread(bhn_auto_reconnect, (void *)minor,
									0);
			if (err < 0) {
				printk("bhn: Couldn't start auto-reconnect thread! (%d)\n", err);
				goto out;
			}
			wait_for_completion(&map->auto_reconn_comp);
			map->auto_reconn_pid = err;
		}
	}
out:
	up(&map->auto_reconn_lock);
}

int prot_udp_connect(struct socket **sock)
{
	int err;
	int on = 1;

	err = ksys_socket(PF_INET, SOCK_DGRAM, 0, sock);
	if (err < 0)
		goto out;

	err = ksys_setsockopt(*sock, SOL_SOCKET, SO_BROADCAST, (char *) &on,
								sizeof(on));

out:
	return err;
}

static int getpacketsize(unsigned char *packet)
{
	struct bhn_cmd *cmd = (struct bhn_cmd *)packet;
	struct bhn_cmd_done *done = (struct bhn_cmd_done *)packet;

	if (cmd->cmd == CMD_DONE)
		return BHN_CMD_DONE_SZ + ntohs(done->length);
	else
		return BHN_CMD_SZ + ntohs(cmd->length);
}

static int partialpacket(unsigned char *cmd, int bytes)
{
	if (((*cmd != CMD_DONE) && (bytes < BHN_CMD_SZ)) ||
	    ((*cmd == CMD_DONE) && (bytes < BHN_CMD_DONE_SZ)) ||
	    (bytes < getpacketsize(cmd)))
		return 1;

	return 0;
}

static unsigned char *getpacket(struct bhn_map *map, unsigned char *cmd)
{
	struct bhn_state *state = map->state;
	unsigned char *buf = state->cmd.read_buf;
	int bytesleft = state->cmd.read_count;

	if (!cmd) {
		/* No previous packet, reset */
		bytesleft = 0;
		cmd = buf;
	} else {
		/* Skip over the previous packet */
		int packet_size = getpacketsize(cmd);
		bytesleft -= packet_size;
		if (bytesleft)
			cmd += packet_size;
		else
			/* All bytes processed, reset */
			cmd = buf;
	}

	while (!bytesleft || partialpacket(cmd, bytesleft)) {
		int buf_size = sizeof(state->cmd.read_buf);
		int err;

		/*
		 * It's possible to have a partial packet that's after a
		 * complete packet. Since we need to read a full packet,
		 * relocate the partial packet to the start of the buffer
		 * to allow maximum free space to read into.
		 */
		if (bytesleft && (cmd != buf)) {
			memmove(buf, cmd, bytesleft);
			cmd = buf;
		}
		down(&map->cr_lock);
		if (state->cmd.sock)
			err = ksys_recv(state->cmd.sock, buf + bytesleft,
						buf_size - bytesleft, 0);
		else
			err = -ENETDOWN;
		up(&map->cr_lock);
		if ((err <= 0) || signal_pending(current)) {
			state->cmd.read_count = 0;
			return NULL;
		}
		bytesleft += err;
	}

	state->cmd.read_count = bytesleft;
	return cmd;
}

int prot_tcp_cmdread(void *arg)
{
	int minor = (int)arg;
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct bhn_prot_cmd_done *done;
	struct bhn_prot_cmd *cmd = NULL;
	struct cmd_state *cstate;

	daemonize();
	snprintf(current->comm, sizeof(current->comm), "bhnS%d cmd", minor);
	current->tty = NULL;
	spin_lock_irq(&current->sigmask_lock);
	siginitsetinv(&current->blocked, sigmask(SIGKILL));
	flush_signals(current);
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);
	reparent_to_init();

	complete(&state->cmd.read_comp);

	for (cmd = (struct bhn_prot_cmd *)getpacket(map,
							(unsigned char *)cmd);
			cmd;
			cmd = (struct bhn_prot_cmd *)getpacket(map,
							(unsigned char *)cmd)) {
		if (cmd->pc_cmd.cmd == CMD_DONE) {
			done = (struct bhn_prot_cmd_done *)cmd;

			spin_lock(&state->listlock);
			list_for_each_entry(cstate, &state->cmdlist, list) {
				if ((done->pr_cmd_done.cmd_completed ==
						cstate->cmd.pc_cmd.cmd) &&
						!is_complete(&cstate->reply))
					break;
			}
			if (&cstate->list != &state->cmdlist) {
				cstate->cmddone = *done;
				complete(&cstate->reply);
			}
			spin_unlock(&state->listlock);

			continue;
		}

		if (cmd->pc_cmd.cmd == EVENT_OCCURRED) {
			if (state->event_cb)
				(state->event_cb)(minor,
						&cmd->pc_event_occurred);
		} else {
			printk("%s: Error: Unknown command '%d'\n", bhn_drvname,
							cmd->pc_cmd.cmd);
		}
	}
	prot_network_not_alive(minor);
	list_for_each_entry(cstate, &state->cmdlist, list) {
		cstate->cmddone.pr_cmd_done.status = CMD_LOCAL_NETDOWN;
		complete(&cstate->reply);
	}
	state->cmd.read_pid = 0;
	complete_and_exit(&state->cmd.read_comp, 0);
}

int prot_tcp_dataread(void *arg)
{
	int minor = (int)arg;
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	unsigned char *ptr;
	int count;
	int err;

	daemonize();
	snprintf(current->comm, sizeof(current->comm), "bhnS%d dr", minor);
	current->tty = NULL;
	spin_lock_irq(&current->sigmask_lock);
	siginitsetinv(&current->blocked, sigmask(SIGKILL));
	flush_signals(current);
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);
	reparent_to_init();

	complete(&state->data.read_comp);

	while (1) {
		down(&map->dr_lock);
		if (state->data.sock)
			err = ksys_recv(state->data.sock, state->data.read_buf,
					sizeof(state->data.read_buf), 0);
		else
			err = -ENETDOWN;
		up(&map->dr_lock);
		if ((err <= 0) || signal_pending(current))
			break;
		count = err;
		ptr = state->data.read_buf;
		do {
			if (map->flags & ASYNC_THROTTLE) {
				wait_event_interruptible(state->wq_throttle,
					!(map->flags & ASYNC_THROTTLE));
				if (signal_pending(current))
					goto loop_out;
			}
			err = receive_chars(minor, ptr, count);
			if (err == count) {
				set_current_state(TASK_INTERRUPTIBLE);
				schedule_timeout((2 * HZ) / 10);
				set_current_state(TASK_RUNNING);
				if (signal_pending(current))
					goto loop_out;
			} else {
				ptr += (count - err);
				count = err;
			}
		} while (count);
	}
loop_out:
	prot_network_not_alive(minor);
	state->data.read_pid = 0;
	complete_and_exit(&state->data.read_comp, 0);
}

int prot_tcp_datawrite(void *arg)
{
	int minor = (int)arg;
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct bhn_connection *data = &state->data;
	struct tty_struct *tty = bhn_driver.table[minor];
	int count;
	int err;

	daemonize();
	snprintf(current->comm, sizeof(current->comm), "bhnS%d dw", minor);
	current->tty = NULL;
	spin_lock_irq(&current->sigmask_lock);
	siginitsetinv(&current->blocked, sigmask(SIGKILL));
	flush_signals(current);
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);
	reparent_to_init();

	complete(&data->write_comp);

	while (1) {
		wait_event_interruptible(data->wq_write,
					data->write_head != data->write_tail);
		if (signal_pending(current))
			break;

		down(&map->auto_reconn_lock);
		if (map->flags & MAP_PURGING) {
			up(&map->auto_reconn_lock);
			wait_event_interruptible(state->wq_purge,
						!(map->flags & MAP_PURGING));
			if (signal_pending(current))
				break;
		} else
			up(&map->auto_reconn_lock);

		count = CIRC_CNT_TO_END(data->write_head, data->write_tail,
								WRITE_BUF_SZ);
		down(&map->dw_lock);
		if (data->sock)
			err = ksys_send(data->sock, data->write_buf +
						data->write_tail, count, 0);
		else
			err = -ENETDOWN;
		up(&map->dw_lock);
		if (err < 0 || signal_pending(current))
			break;

		data->write_tail = (data->write_tail + count) % WRITE_BUF_SZ;

		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
							tty->ldisc.write_wakeup)
			(tty->ldisc.write_wakeup)(tty);
		wake_up(&tty->write_wait);
#ifdef SERIAL_HAVE_POLL_WAIT
		wake_up(&tty->poll_wait);
#endif
	}
	prot_network_not_alive(minor);
	state->data.write_pid = 0;
	complete_and_exit(&data->write_comp, 0);
}

int prot_ping(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_unit *unit = map->unit;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	struct socket *sock;
	int timeout;
	int err;

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = PING;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = 0;
	cmd->pc_cmd.length = 0;

	if (map->flags & MAP_AUTO_RECONNECT) {
		timeout = map->cmd_timeout / HZ;
		if (!timeout)
			timeout = 1;
	} else
		timeout = 0;

	err = prot_tcp_connect(&sock, (SOCKADDR)&unit->addr, timeout);
	if (err < 0)
		goto free_out;

	err = ksys_send(sock, cmd, PING_CMD_SZ, 0);
	if (err < 0)
		goto sock_out;
	if (err != PING_CMD_SZ) {
		err = -EIO;
		goto sock_out;
	}

	cmddone = &cstate->cmddone;
	err = ksys_recv(sock, cmddone, sizeof(*cmddone), 0);
	if (err < 0)
		goto sock_out;
	if (err == 0) {
		err = -ENETDOWN;
		goto sock_out;
	}

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto sock_out;

	err = 0;

sock_out:
	sock_release(sock);
free_out:
	kfree(cstate);
out:
	return err;
}

int prot_open(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int timeout;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = OPEN_PORT;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = htons(BHN_OPEN_SZ);
	cmd->pc_open.flags = OPEN_RDWR;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, OPEN_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != OPEN_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		err = -EIO;
		map->flags &= ~MAP_NETWORK_ALIVE;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;

	/*
	 * A successful reconnection means that the port on the firmware
	 * has closed and re-opened, losing its settings. Compensate.
	 */
	down(&map->auto_reconn_lock);
	if (map->auto_reconn_pid)
		bhn_do_change_speed(minor);
	up(&map->auto_reconn_lock);

	if (map->flags & MAP_AUTO_RECONNECT) {
		timeout = map->cmd_timeout / HZ;
		if (!timeout)
			timeout = 1;
	} else
		timeout = 0;

	down(&map->dr_lock);
	down(&map->dw_lock);
	err = prot_tcp_connect(&state->data.sock, (SOCKADDR)&state->data.addr,
								timeout);
	up(&map->dw_lock);
	up(&map->dr_lock);
	if (err < 0)
		goto list_out;

	/*
	 * If we don't wait for the threads to start up, it's possible that
	 * the kill_proc() can happen before the threads have started, and
	 * the SIGKILL gets lost resulting in hung user space apps. This
	 * happens in practice.
	 */
	err = kernel_thread(prot_tcp_dataread, (void *)minor, 0);
	if (err < 0)
		goto connect_out;
	wait_for_completion(&state->data.read_comp);
	state->data.read_pid = err;

	err = kernel_thread(prot_tcp_datawrite, (void *)minor, 0);
	if (err < 0)
		goto thread_out;
	wait_for_completion(&state->data.write_comp);
	state->data.write_pid = err;

	map->flags |= MAP_OPENED;
	err = 0;
	goto list_out;

thread_out:
	kill_proc(state->data.read_pid, SIGKILL, 1);
	wait_for_completion(&state->data.read_comp);
connect_out:
	down(&map->dr_lock);
	down(&map->dw_lock);
	sock_release(state->data.sock);
	up(&map->dw_lock);
	up(&map->dr_lock);
list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_close(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err = 0;

	if (!state)
		goto out;

	if (!(map->flags & MAP_OPENED))
		goto out;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto thread_out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto thread_out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = CLOSE_PORT;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = 0;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, CLOSE_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != CLOSE_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		map->flags &= ~MAP_NETWORK_ALIVE;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	if (cmddone->pr_cmd_done.status != CMD_OK)
		err = -EPROTO;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);

thread_out:
	if (state->data.read_pid) {
		kill_proc(state->data.read_pid, SIGKILL, 1);
		wait_for_completion(&state->data.read_comp);
	}

	if (state->data.write_pid) {
		kill_proc(state->data.write_pid, SIGKILL, 1);
		wait_for_completion(&state->data.write_comp);
	}

	down(&map->dr_lock);
	down(&map->dw_lock);
	sock_release(state->data.sock);
	up(&map->dw_lock);
	up(&map->dr_lock);

	map->flags &= ~MAP_OPENED;

out:
	return err;
}

int prot_query_tx_finished(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;
	int err_count = 0;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	do {
		init_completion(&cstate->reply);

		cmd = &cstate->cmd;
		cmd->pc_cmd.cmd = QUERY_TX_FINISHED;
		cmd->pc_cmd.sequence = 0;
		cmd->pc_cmd.port = map->unit_port;
		cmd->pc_cmd.length = 0;

		spin_lock(&state->listlock);
		list_add_tail(&cstate->list, &state->cmdlist);
		spin_unlock(&state->listlock);

		down(&map->cw_lock);
		if (state->cmd.sock)
			err = ksys_send(state->cmd.sock, cmd,
						QUERY_TX_FINISHED_CMD_SZ, 0);
		else
			err = -ENETDOWN;
		up(&map->cw_lock);
		if (err < 0)
			goto list_out;
		if (err != QUERY_TX_FINISHED_CMD_SZ) {
			err = -EIO;
			goto list_out;
		}

		err = wait_for_completion_interruptible_timeout(&cstate->reply,
					map->cmd_timeout);
		if (err < 0)
			goto list_out;
		if (!err) {
			prot_network_not_alive(minor);
			err = -ENETDOWN;
			goto list_out;
		}

		cmddone = &cstate->cmddone;

		err = -EPROTO;
		if (cmddone->pr_cmd_done.status == CMD_LOCAL_NETDOWN)
			err = -ENETDOWN;
		else if (cmddone->pr_cmd_done.status != CMD_OK)
			err_count++;
		else
			err = 0;

		/*
		 * Currently we only wait for the good response. The event
		 * callback routine is responsible for actually waking up
		 * waiting thread when the TX_FIN event arrives. This allows
		 * us to handle the condition where the event might arrive
		 * before the command response, which is quite often in
		 * practice, due to the way the scheduler in the firmware is.
		 */

list_out:
		spin_lock(&state->listlock);
		list_del(&cstate->list);
		spin_unlock(&state->listlock);
	}
	while(err_count < 10 && err == -EPROTO);

	kfree(cstate);
out:
	return err;
}

int prot_purge(int minor, int flags)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	down(&map->auto_reconn_lock);
	map->flags |= MAP_PURGING;
	up(&map->auto_reconn_lock);

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = PURGE;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = htons(BHN_PURGE_SZ);
	cmd->pc_purge.flags = flags;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, PURGE_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != PURGE_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	down(&map->auto_reconn_lock);
	map->flags &= ~MAP_PURGING;
	up(&map->auto_reconn_lock);
	wake_up_interruptible(&state->wq_throttle);
	return err;
}

int prot_set_line(int minor, struct tty_struct *tty)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;
	unsigned int cflag = tty->termios->c_cflag;
	unsigned int iflag = tty->termios->c_iflag;
	unsigned char *cc = tty->termios->c_cc;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = SET_LINE;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = htons(BHN_LSET_SZ);

	memset(&cmd->pc_lset, 0, BHN_LSET_SZ);
	cmd->pc_lset.baud = htonl(tty_get_baud_rate(tty));
	switch (cflag & CSIZE) {
	case CS5: cmd->pc_lset.databits = 5; break;
	case CS6: cmd->pc_lset.databits = 6; break;
	case CS7: cmd->pc_lset.databits = 7; break;
	case CS8: cmd->pc_lset.databits = 8; break;
	}
	if (!(cflag & PARENB))
		cmd->pc_lset.parity = LSET_PARITY_NONE;
	else if ((cflag & PARODD) == PARODD) {
		if ((cflag & CMSPAR) == CMSPAR)
			cmd->pc_lset.parity = LSET_PARITY_MARK;
		else
			cmd->pc_lset.parity = LSET_PARITY_ODD;
	} else {
		if ((cflag & CMSPAR) == CMSPAR)
			cmd->pc_lset.parity = LSET_PARITY_SPACE;
		else
			cmd->pc_lset.parity = LSET_PARITY_EVEN;
	}
	if ((cflag & CSTOPB) == CSTOPB)
		cmd->pc_lset.stopbits = 2;
	else
		cmd->pc_lset.stopbits = 1;
	if ((iflag & (IXON | IXOFF)) == (IXON | IXOFF))
		cmd->pc_lset.sflow = LSET_SFLOW_RXTX;
	else if ((iflag & IXON) == IXON)
		cmd->pc_lset.sflow = LSET_SFLOW_RX;
	else if ((iflag & IXOFF) == IXOFF)
		cmd->pc_lset.sflow = LSET_SFLOW_TX;
	else
		cmd->pc_lset.sflow = LSET_SFLOW_NONE;
	cmd->pc_lset.xoff = cc[VSTOP];
	cmd->pc_lset.xon = cc[VSTART];
	if ((cflag & CRTSCTS) == CRTSCTS)
		cmd->pc_lset.hflow = LSET_HFLOW_RTS | LSET_HFLOW_CTS;
	else
		cmd->pc_lset.hflow = LSET_HFLOW_NONE;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, SET_LINE_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != SET_LINE_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_get_line(int minor, struct tty_struct *tty)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;
	unsigned int cflag = tty->termios->c_cflag;
	unsigned int iflag = tty->termios->c_iflag;
	int i;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = GET_LINE;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = 0;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, GET_LINE_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != GET_LINE_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;
	if (ntohs(cmddone->pr_cmd_done.length) != BHN_LSET_SZ)
		goto list_out;

	/* The following should be the inverse of tty_get_baud_rate() */
	cflag &= ~(CBAUD | CBAUDEX);
	i = tty_get_baud_idx(cmddone->pr_lset.baud);
	if (i == -1)
		goto list_out;
	cflag |= i;
	cflag &= ~CSIZE;
	switch (cmddone->pr_lset.databits) {
	case 5:
		cflag |= CS5;
		break;
	case 6:
		cflag |= CS6;
		break;
	case 7:
		cflag |= CS7;
		break;
	case 8:
		cflag |= CS8;
		break;
	default:
		goto list_out;
	}
	if (cmddone->pr_lset.parity == LSET_PARITY_NONE)
		cflag &= ~PARENB;
	else {
		cflag |= PARENB;
		if (cmddone->pr_lset.parity == LSET_PARITY_EVEN) {
			cflag &= ~CMSPAR;
			cflag &= ~PARODD;
		} else if (cmddone->pr_lset.parity == LSET_PARITY_ODD) {
			cflag &= ~CMSPAR;
			cflag |= PARODD;
		} else if (cmddone->pr_lset.parity == LSET_PARITY_SPACE) {
			cflag |= CMSPAR;
			cflag &= ~PARODD;
		} else if (cmddone->pr_lset.parity == LSET_PARITY_MARK) {
			cflag |= CMSPAR;
			cflag |= PARODD;
		} else
			goto list_out;
	}
	if (cmddone->pr_lset.stopbits == 1)
		cflag &= ~CSTOPB;
	else
		cflag |= CSTOPB;
	if (cmddone->pr_lset.sflow == LSET_SFLOW_NONE)
		iflag &= ~(IXON | IXOFF | IXANY);
	else if (cmddone->pr_lset.sflow == LSET_SFLOW_RX)
		iflag |= IXON;
	else if (cmddone->pr_lset.sflow == LSET_SFLOW_TX)
		iflag |= IXOFF;
	else if (cmddone->pr_lset.sflow == LSET_SFLOW_RXTX)
		iflag |= (IXON | IXOFF);
	else
		goto list_out;
	if (cmddone->pr_lset.hflow == LSET_HFLOW_NONE)
		cflag &= ~CRTSCTS;
	else {
		if (cmddone->pr_lset.hflow & LSET_HFLOW_RTS)
			cflag |= CRTSCTS;
		if (cmddone->pr_lset.hflow & LSET_HFLOW_CTS)
			cflag |= CRTSCTS;
	}
	tty->termios->c_cflag = cflag;
	tty->termios->c_iflag = iflag;
	tty->termios->c_cc[VSTOP] = cmddone->pr_lset.xoff;
	tty->termios->c_cc[VSTART] = cmddone->pr_lset.xon;
	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_set_lim(int minor, int mode)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = SET_LIM;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = htons(BHN_LIM_SZ);
	cmd->pc_lim.mode = mode;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, SET_LIM_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != SET_LIM_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_get_lim(int minor, int *mode)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = GET_LIM;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = 0;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, GET_LIM_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != GET_LIM_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;
	if (ntohs(cmddone->pr_cmd_done.length) != BHN_LIM_SZ)
		goto list_out;

	*mode = cmddone->pr_lim.mode;

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_get_line_state(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = GET_LINE_STATE;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = 0;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, GET_LINE_STATE_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != GET_LINE_STATE_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;
	if (ntohs(cmddone->pr_cmd_done.length) != BHN_LSTATE_SZ)
		goto list_out;

	/* do something with the line state */

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_send_break(int minor, int time)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = SEND_BREAK;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = htons(BHN_BREAK_SZ);
	cmd->pc_break.duration = time;

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, SEND_BREAK_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != SEND_BREAK_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = 0;

list_out:
	kfree(cstate);
out:
	return err;
}

int prot_start_stop_break(int minor, int onoff)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = START_STOP_BREAK;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = htons(BHN_SSBREAK_SZ);
	cmd->pc_ssbreak.onoff = onoff;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, START_STOP_BREAK_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != START_STOP_BREAK_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_set_rts(int minor, int onoff)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = SET_RTS;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = htons(BHN_RTSDTR_SZ);
	cmd->pc_rtsdtr.onoff = onoff;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, SET_RTS_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != SET_RTS_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_set_dtr(int minor, int onoff)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = SET_DTR;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = htons(BHN_RTSDTR_SZ);
	cmd->pc_rtsdtr.onoff = onoff;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, SET_DTR_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != SET_DTR_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_resume_tx(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = RESUME_TX;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = 0;

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, RESUME_TX_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != RESUME_TX_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = 0;

list_out:
	kfree(cstate);
out:
	return err;
}

int prot_host_type(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = HOST_TYPE;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = 0;

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, HOST_TYPE_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != HOST_TYPE_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = 0;

list_out:
	kfree(cstate);
out:
	return err;
}

int prot_get_cds(int minor, struct bhn_cds_parameters *cds_para)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_unit *unit = map->unit;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	struct socket *sock;
	int timeout;
	int err;
	int nbytes;

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = GET_CDS;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = 0;
	cmd->pc_cmd.length = 0;

	if (map->flags & MAP_AUTO_RECONNECT) {
		timeout = map->cmd_timeout / HZ;
		if (!timeout)
			timeout = 1;
	} else
		timeout = 0;

	err = prot_tcp_connect(&sock, (SOCKADDR)&unit->addr, timeout);
	if (err < 0)
		goto free_out;

	err = ksys_send(sock, cmd, GET_CDS_CMD_SZ, 0);
	if (err < 0)
		goto sock_out;
	if (err != GET_CDS_CMD_SZ) {
		err = -EIO;
		goto sock_out;
	}

	cmddone = &cstate->cmddone;
	nbytes = 0;
	do {
		err = ksys_recv(sock, ((unsigned char *)cmddone + nbytes),
						sizeof(*cmddone) - nbytes, 0);
		if (err < 0)
			goto sock_out;
		if (err == 0) {
			err = -ENETDOWN;
			goto sock_out;
		}
		nbytes += err;
	} while (nbytes < GET_CDS_CMD_DONE_SZ);

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto sock_out;

	*cds_para = cmddone->pr_cdsp;

	err = 0;

sock_out:
	sock_release(sock);
free_out:
	kfree(cstate);
out:
	return err;
}

/*
 * CONNECT_PORT_CMD - normal command to a specific port, get a specific reply
 * CONNECT_PORT_GEN - same as above, but to the general command port
 */
int prot_get_bhn_info(int minor, int how_con, struct bhn_info *bhn_info,
				unsigned char *mac, unsigned char *hostname)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_unit *unit = map->unit;
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	struct socket *sock;
	int timeout;
	int err;

	switch (how_con) {
		case CONNECT_PORT_CMD:
			if (!(map->flags & MAP_NETWORK_ALIVE)) {
				err = -ENETDOWN;
				goto out;
			}
			break;
		case CONNECT_PORT_GEN:
			break;
		default:
			err = -EINVAL;
			goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	if (how_con == CONNECT_PORT_CMD)
		init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = GET_BHN_INFO;
	cmd->pc_cmd.sequence = 0;
	if (how_con == CONNECT_PORT_CMD)
		cmd->pc_cmd.port = map->unit_port;
	else
		cmd->pc_cmd.port = 0;
	cmd->pc_cmd.length = htons(BHN_INFO_CMD_SZ);
	memset(&cmd->pc_info_cmd, 0, BHN_INFO_CMD_SZ);
	if (mac)
		memcpy(cmd->pc_info_cmd.bhn_mac, mac, 6);
	if (hostname)
		memcpy(cmd->pc_info_cmd.bhn_hostname, hostname, HOST_NAME_SIZE);

	if (how_con == CONNECT_PORT_CMD) {
		spin_lock(&state->listlock);
		list_add_tail(&cstate->list, &state->cmdlist);
		spin_unlock(&state->listlock);
		down(&map->cw_lock);
		sock = state->cmd.sock;
	} else {
		if (map->flags & MAP_AUTO_RECONNECT) {
			timeout = map->cmd_timeout / HZ;
			if (!timeout)
				timeout = 1;
		} else
			timeout = 0;

		err = prot_tcp_connect(&sock, (SOCKADDR)&unit->addr, timeout);
		if (err < 0)
			goto free_out;
	}

	if (sock)
		err = ksys_send(sock, cmd, GET_BHN_INFO_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	if (how_con == CONNECT_PORT_CMD)
		up(&map->cw_lock);
	if (err < 0)
		goto cmd_out;
	if (err != GET_BHN_INFO_CMD_SZ) {
		err = -EIO;
		goto cmd_out;
	}

	cmddone = &cstate->cmddone;
	if (how_con == CONNECT_PORT_CMD) {
		err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
		if (err < 0)
			goto cmd_out;
		if (!err) {
			prot_network_not_alive(minor);
			err = -ENETDOWN;
			goto cmd_out;
		}
	} else {
		err = ksys_recv(sock, cmddone, sizeof(*cmddone), 0);
		if (err < 0)
			goto cmd_out;
		if (err == 0) {
			err = -ENETDOWN;
			goto cmd_out;
		}
	}

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto cmd_out;

	err = 0;

	*bhn_info = cmddone->pr_binfo;

cmd_out:
	if (how_con == CONNECT_PORT_CMD) {
		spin_lock(&state->listlock);
		list_del(&cstate->list);
		spin_unlock(&state->listlock);
	} else
		sock_release(sock);
free_out:
	kfree(cstate);
out:
	return err;
}

int prot_get_port_info(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = GET_PORT_INFO;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = 0;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, GET_PORT_INFO_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != GET_PORT_INFO_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;
	if (ntohs(cmddone->pr_cmd_done.length) != BHN_PORTINFO_SZ)
		goto list_out;

	/* do something with portinfo data */

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_set_connection(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = SET_CONNECTION;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = htons(BHN_CSET_SZ);
	/* init the struct with some data */

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, SET_CONNECTION_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != SET_CONNECTION_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_fatal_error(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = FATAL_ERROR;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = 0;

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, FATAL_ERROR_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != FATAL_ERROR_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = 0;

list_out:
	kfree(cstate);
out:
	return err;
}

int prot_port_reset(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = PORT_RESET;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = 0;

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, PORT_RESET_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != PORT_RESET_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = 0;

list_out:
	kfree(cstate);
out:
	return err;
}

int prot_reset_bhn(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = RESET_BHN;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = 0;

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, RESET_BHN_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != RESET_BHN_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = 0;

list_out:
	kfree(cstate);
out:
	return err;
}

int prot_send_immediate(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = SEND_IMMEDIATE;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = htons(BHN_SENDI_SZ);
	/* init the data */

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, SEND_IMMEDIATE_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != SEND_IMMEDIATE_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = 0;

list_out:
	kfree(cstate);
out:
	return err;
}

int prot_fifo_control(int minor, int rx, int tx)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = FIFO_CONTROL;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = htons(BHN_FIFO_SZ);
	cmd->pc_fifo.rx = rx;
	cmd->pc_fifo.tx = tx;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, FIFO_CONTROL_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != FIFO_CONTROL_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;
	if (ntohs(cmddone->pr_cmd_done.length) != BHN_FIFO_SZ)
		goto list_out;

	/* do something with the returned data */

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_set_event(int minor, int events)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = SET_EVENT;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = htons(BHN_EVENT_SZ);
	cmd->pc_event.events = htons(events);

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, SET_EVENT_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != SET_EVENT_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_get_event(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = GET_EVENT;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = 0;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, GET_EVENT_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != GET_EVENT_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;
	if (ntohs(cmddone->pr_cmd_done.length) != BHN_EVENT_SZ)
		goto list_out;

	/* do something with the data */

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_get_msr(int minor, unsigned char *msr)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = GET_MSR;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = 0;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, GET_MSR_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != GET_MSR_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;
	if (ntohs(cmddone->pr_cmd_done.length) != BHN_MSR_SZ)
		goto list_out;

	*msr = cmddone->pr_msr.msr;

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_send_xchar(int minor, char xchar)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = SEND_XCHAR;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = htons(BHN_XCHAR_SZ);
	cmd->pc_xchar.xchar = xchar;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, SEND_XCHAR_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != SEND_XCHAR_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_set_rts_dtr(int minor, int rts_onoff, int dtr_onoff)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = SET_RTS_DTR;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = htons(BHN_RTSANDDTR_SZ);
	cmd->pc_rtsanddtr.rts_onoff = rts_onoff ? RTS_ON : RTS_OFF;
	cmd->pc_rtsanddtr.dtr_onoff = dtr_onoff ? DTR_ON : DTR_OFF;
	/* init data */

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, SET_RTS_DTR_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != SET_RTS_DTR_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_get_portcap(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = GET_PORTCAP;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = 0;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, GET_PORTCAP_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != GET_PORTCAP_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;
	if (ntohs(cmddone->pr_cmd_done.length) != BHN_PORTCAP_SZ)
		goto list_out;

	/* do something with the data */

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

int prot_get_mfginfo(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	if (!(map->flags & MAP_NETWORK_ALIVE)) {
		err = -ENETDOWN;
		goto out;
	}

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = GET_MFGINFO;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = 0;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, GET_MFGINFO_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != GET_MFGINFO_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;
	if (ntohs(cmddone->pr_cmd_done.length) != BHN_MFGINFO_SZ)
		goto list_out;

	/* do something with the data */

	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

static struct cmd_state autod_cstate;
static struct socket *autod_sock;

int prot_start_autod(int port, unsigned int netaddr, unsigned char *mac,
							unsigned char *hostname)
{
	struct bhn_prot_cmd *cmd;
	struct sockaddr_in addr;
	int err;

	cmd = &autod_cstate.cmd;
	cmd->pc_cmd.cmd = GET_BHN_INFO;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = 0;
	cmd->pc_cmd.length = htons(BHN_INFO_CMD_SZ);
	memset(&cmd->pc_info_cmd, 0, BHN_INFO_CMD_SZ);
	if (mac)
		memcpy(cmd->pc_info_cmd.bhn_mac, mac, 6);
	if (hostname)
		memcpy(cmd->pc_info_cmd.bhn_hostname, hostname, HOST_NAME_SIZE);

	addr.sin_family = PF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = netaddr;
	err = prot_udp_connect(&autod_sock);
	if (err < 0)
		goto out;

	err = ksys_sendto(autod_sock, cmd, GET_BHN_INFO_CMD_SZ, 0,
						(SOCKADDR)&addr, sizeof(addr));
	if (err < 0)
		goto sock_out;
	if (err != GET_BHN_INFO_CMD_SZ) {
		err = -EIO;
		goto sock_out;
	}

	err = 0;
	goto out;

sock_out:
	sock_release(autod_sock);
out:
	return err;
}

int prot_get_autod(struct bhn_info *info, long poll, long timeout)
{
	struct bhn_prot_cmd_done *cmddone = &autod_cstate.cmddone;
	int flags = 0;
	int err;

	if (timeout)
		flags = MSG_DONTWAIT;

retry:
	err = ksys_recv(autod_sock, cmddone, sizeof(*cmddone), flags);
	if (err < 0) {
		if ((err == -EAGAIN) && timeout) {
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(poll);
			set_current_state(TASK_RUNNING);
			if (signal_pending(current)) {
				err = -EINTR;
				goto out;
			}
			timeout -= poll;
			if (timeout < 0)
				timeout = 0;
			goto retry;
		}
		goto out;
	}
	if (err == 0) {
		err = -ENETDOWN;
		goto out;
	}

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto out;

	err = 0;
	*info = cmddone->pr_binfo;

out:
	return err;
}

void prot_stop_autod(void)
{
	struct bhn_prot_cmd_done *cmddone = &autod_cstate.cmddone;
	int err;

	while (1) {
		err = ksys_recv(autod_sock, cmddone, sizeof(*cmddone),
								MSG_DONTWAIT);
		if (err <= 0)
			break;
	}

	sock_release(autod_sock);
}

#if 0
int prot_xxxx(int minor)
{
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct cmd_state *cstate;
	struct bhn_prot_cmd *cmd;
	struct bhn_prot_cmd_done *cmddone;
	int err;

	cstate = kmalloc(sizeof(*cstate), GFP_KERNEL);
	if (!cstate) {
		err = -ENOMEM;
		goto out;
	}
	init_completion(&cstate->reply);

	cmd = &cstate->cmd;
	cmd->pc_cmd.cmd = yyyy;
	cmd->pc_cmd.sequence = 0;
	cmd->pc_cmd.port = map->unit_port;
	cmd->pc_cmd.length = zzzz;
	aaaa;

	spin_lock(&state->listlock);
	list_add_tail(&cstate->list, &state->cmdlist);
	spin_unlock(&state->listlock);

	down(&map->cw_lock);
	if (state->cmd.sock)
		err = ksys_send(state->cmd.sock, cmd, yyyy_CMD_SZ, 0);
	else
		err = -ENETDOWN;
	up(&map->cw_lock);
	if (err < 0)
		goto list_out;
	if (err != yyyy_CMD_SZ) {
		err = -EIO;
		goto list_out;
	}

	err = wait_for_completion_interruptible_timeout(&cstate->reply,
							map->cmd_timeout);
	if (err < 0)
		goto list_out;
	if (!err) {
		prot_network_not_alive(minor);
		err = -ENETDOWN;
		goto list_out;
	}

	cmddone = &cstate->cmddone;

	err = -EPROTO;
	if (cmddone->pr_cmd_done.status != CMD_OK)
		goto list_out;

	bbbb;
	err = 0;

list_out:
	spin_lock(&state->listlock);
	list_del(&cstate->list);
	spin_unlock(&state->listlock);
	kfree(cstate);
out:
	return err;
}

#endif
