/*
 *
 *  The Blue Heat/Net driver
 *
 *  Copyright (C) 2004 Connect Tech Inc.
 *
 *  Apr 28 2004 Stuart MacDonald <stuartm@connecttech.com>
 *    First draft of the driver.
 *
 *  Sep 28 2004 Stuart MacDonald <stuartm@connecttech.com>
 *    Second draft with direct-to-TCP code.
 *
 *  Nov 28 2004 Stuart MacDonald <stuartm@connecttech.com>
 *    Bug fixes for data tx and rx.
 *
 */

static char *bhn_version = "2.16";
static char *bhn_revdate = "Apr 27 2007";
char *bhn_drvname = "BlueHeatNet";

#undef USE_KDB_DEBUG

#if 0
#define SERIAL_DEBUG_INTR
#define SERIAL_DEBUG_OPEN
#define SERIAL_DEBUG_FLOW
#define SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
#define SERIAL_DEBUG_PCI
#define SERIAL_DEBUG_AUTOCONF
#define SERIAL_DEBUG_IOCTL
#define SERIAL_DEBUG_READ
#define SERIAL_DEBUG_WRITE
#define SERIAL_DEBUG_START
#define SERIAL_DEBUG_OPENLOCK
#endif

#ifdef SERIAL_DEBUG_OPENLOCK
#define BHNDEBUG_OPENLOCK(_args_) \
        do { printk _args_;} while(0)
#else
#define BHNDEBUG_OPENLOCK(_args_) \
	do { } while(0)
#endif

/* borrowed to gate calls to prot_query_tx_finished */
#define	ASYNC_QUERY_IN_PROGRESS	0x00000100

#include <linux/version.h>
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#if (LINUX_VERSION_CODE >= 0x02010f)
#include <linux/init.h>
#endif
#include <linux/types.h>
#include <asm/serial.h>
#include <linux/serial.h>
#include <linux/serialP.h>
#include <linux/serial_reg.h>
#include <linux/fs.h>
#include <linux/tty.h>
#include <net/sock.h>
#include <asm/uaccess.h>
#include <asm/delay.h>
#include <linux/proc_fs.h>
#ifdef USE_KDB_DEBUG
#include <linux/kdb.h>
#include <linux/kdbprivate.h>
#endif
#include <linux/list.h>
#include <linux/ctype.h>
#include <asm/irq.h>
#include <linux/completion.h>

#include "ksocket.h"

#include <BHN_flash.h>
#include <CDS.h>
#include <BHN_protocol.h>
#include "bhn.h"
#define NEED_TTY_LDISC_REF
#define NEED_TTY_LDISC_DEREF
#define NEED_SSCANF
#include "bhn_compat.h"
#include "bhn_elist.h"

#define DEFAULT_CMD_TIMEOUT	(300)
#define DEFAULT_AUTO_RECONNECT	(0)

#if defined(MODULE) && defined(SERIAL_DEBUG_MCOUNT)
#define DBG_CNT(s) printk("(%s): [%x] refc=%d, serc=%d, ttyc=%d -> %s\n", \
				kdevname(tty->device), (info->flags),	\
				serial_refcount, info->count,	\
				tty->count, (s))
#else
#define DBG_CNT(s)
#endif

#define SKIP_WSPACE(p)		while (isspace(*(p))) (p)++
#define SKIP_TOKEN(p)		while (isgraph(*(p)) && (*(p) != '=')) (p)++
#define SKIPN_LINE(p, n)	while ((n) && (*(p) != '\n')) { (p)++; (n)--; }
#define SKIP_COMMENT(p)		if (*(p) == '#') continue
#define SKIP_EMPTY(p)		if (*(p) == '\0') continue
#define EQUALS(p)		if (*(p) == '=') (p)++; else return -EINVAL
#define ASSIGN_OR_EOB(p, x)	if (*(p)) x = (p); else break
#define ASSIGN(p, x)		if (*(p)) x = (p); else return -EINVAL

#define AUTOD_FORMAT	"%3d %d.%d.%d.%d %5d %2d %d %d\n"

#define MAP_TO_MINOR(map)	((map) - bhn_table)

struct serial_state rs_table[NPORTS];
unsigned char uart_msr[NPORTS];
unsigned char uart_lsr[NPORTS];

/*
 * TTY functions
 */

static int bhn_open(struct tty_struct *tty, struct file *filp);
static void bhn_close(struct tty_struct *tty, struct file *filp);
static int bhn_write(struct tty_struct *tty, int from_user,
					const unsigned char *buf, int count);
static void bhn_flush_chars(struct tty_struct *tty);
static int bhn_write_room(struct tty_struct *tty);
static int bhn_chars_in_buffer(struct tty_struct *tty);
static int bhn_ioctl(struct tty_struct *tty, struct file *file,
					unsigned int cmd, unsigned long arg);
static void bhn_set_termios(struct tty_struct *tty,
						struct termios *old_term);
static void bhn_throttle(struct tty_struct *tty);
static void bhn_unthrottle(struct tty_struct *tty);
static void bhn_stop(struct tty_struct *tty);
static void bhn_start(struct tty_struct *tty);
static void bhn_hangup(struct tty_struct *tty);
#if (LINUX_VERSION_CODE >= 0x020142)
static void bhn_break(struct tty_struct *tty, int onoff);
#endif
static void bhn_flush_buffer(struct tty_struct *tty);
#if (LINUX_VERSION_CODE >= 0x02010f)
static void bhn_wait_until_sent(struct tty_struct *tty, int timeout);
static void bhn_send_xchar(struct tty_struct *tty, char c);
#endif
#ifdef CONFIG_PROC_FS
static int bhn_read_proc(char *page, char **start, off_t off, int count,
							int *eof, void *data);
static int bhn_write_proc(struct file *file, const char *buffer,
					unsigned long count, void *data);
static int bhn_proc_autod_read(char *page, char **start, off_t off, int count,
							int *eof, void *data);
static int bhn_proc_autod_write(struct file *file, const char *buffer,
					unsigned long count, void *data);
static int bhn_proc_config_read(char *page, char **start, off_t off, int count,
							int *eof, void *data);
static int bhn_proc_config_write(struct file *file, const char *buffer,
					unsigned long count, void *data);
#endif

/*
 * Helper functions
 */

static void change_speed(struct async_struct *info, struct termios *old);
static void event_handler(int minor, struct bhn_event_occurred *e);

/*
 * Global variables
 */

static struct tty_driver bhn_driver;
static struct tty_driver bhn_callout_driver;

struct bhn_unit bhn_unit_table[NPORTS];
struct bhn_map bhn_table[NPORTS];

static struct tty_struct *bhn_tty[NPORTS];
static struct termios *bhn_termios[NPORTS];
static struct termios *bhn_termios_locked[NPORTS];

static int bhn_refcount;

#define PROC_BHNROOT_NM		"bhn"
#define PROC_AUTOD_NM		"autodiscover"
#define PROC_CONFIG_NM		"config"
#define PROC_STATS_NM		"stats"

static struct proc_dir_entry *bhn_proc_root;
static struct proc_dir_entry *bhn_proc_autodiscover;
static struct proc_dir_entry *bhn_proc_config;
static struct proc_dir_entry *bhn_proc_stats;

#define AUTOD_NETWORK_NM	"network"
#define AUTOD_PORT_NM		"port"
#define AUTOD_TIMEOUT_NM	"timeout"
#define AUTOD_POLL_NM		"poll"
#define AUTOD_NETWORK_SZ	strlen(AUTOD_NETWORK_NM)
#define AUTOD_PORT_SZ		strlen(AUTOD_PORT_NM)
#define AUTOD_TIMEOUT_SZ	strlen(AUTOD_TIMEOUT_NM)
#define AUTOD_POLL_SZ		strlen(AUTOD_POLL_NM)

static unsigned char autod_netaddr[4] = { 255, 255, 255, 255 };
static int autod_port = BHN_DEF_BASE_PORT;
static int autod_timeout = 50;
static int autod_poll = 1;

/*
 * Serial Functions
 */

static int block_til_ready(struct tty_struct *tty, struct file *filp,
			   struct async_struct *info)
{
	DECLARE_WAITQUEUE(wait, current);
	struct serial_state *state = info->state;
	int retval;
	int do_clocal = 0, extra_count = 0;
	unsigned long flags;

	/*
	 * If the device is in the middle of being closed, then block
	 * until it's done, and then try again.
	 */
	if (tty_hung_up_p(filp) || (info->flags & ASYNC_CLOSING)) {
		if (info->flags & ASYNC_CLOSING)
			interruptible_sleep_on(&info->close_wait);
#ifdef SERIAL_DO_RESTART
		return ((info->flags & ASYNC_HUP_NOTIFY) ?
			-EAGAIN : -ERESTARTSYS);
#else
		return -EAGAIN;
#endif
	}

	/*
	 * If this is a callout device, then just make sure the normal
	 * device isn't being used.
	 */
	if (tty->driver.subtype == SERIAL_TYPE_CALLOUT) {
		if (info->flags & ASYNC_NORMAL_ACTIVE) {
			return -EBUSY;
		}
		if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    (info->flags & ASYNC_SESSION_LOCKOUT) &&
		    (info->session != current->session)) {
			return -EBUSY;
		}
		if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    (info->flags & ASYNC_PGRP_LOCKOUT) &&
		    (info->pgrp != current->pgrp))
			return -EBUSY;
		info->flags |= ASYNC_CALLOUT_ACTIVE;
		return 0;
	}

	/*
	 * If non-blocking mode is set, or the port is not enabled,
	 * then make the check up front and then exit.
	 */
	if ((filp->f_flags & O_NONBLOCK) ||
	    (tty->flags & (1 << TTY_IO_ERROR))) {
		if (info->flags & ASYNC_CALLOUT_ACTIVE)
			return -EBUSY;
		info->flags |= ASYNC_NORMAL_ACTIVE;
		return 0;
	}

	if (info->flags & ASYNC_CALLOUT_ACTIVE) {
		if (state->normal_termios.c_cflag & CLOCAL)
			do_clocal = 1;
	} else {
		if (tty->termios->c_cflag & CLOCAL)
			do_clocal = 1;
	}

	/*
	 * Block waiting for the carrier detect and the line to become
	 * free (i.e., not in use by the callout).  While we are in
	 * this loop, state->count is dropped by one, so that
	 * bhn_close() knows when to free things.  We restore it upon
	 * exit, either normal or abnormal.
	 */
	retval = 0;
	add_wait_queue(&info->open_wait, &wait);
#ifdef SERIAL_DEBUG_OPEN
	printk("%s - before block: bhnS%d, count = %d\n", __FUNCTION__,
						state->line, state->count);
#endif
	save_flags(flags); cli();
	if (!tty_hung_up_p(filp)) {
		extra_count = 1;
		state->count--;
	}
	restore_flags(flags);
	info->blocked_open++;
	while (1) {
		set_current_state(TASK_INTERRUPTIBLE);
		if (tty_hung_up_p(filp) ||
		    !(info->flags & ASYNC_INITIALIZED)) {
#ifdef SERIAL_DO_RESTART
			if (info->flags & ASYNC_HUP_NOTIFY)
				retval = -EAGAIN;
			else
				retval = -ERESTARTSYS;
#else
			retval = -EAGAIN;
#endif
			break;
		}
		if (!(info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    !(info->flags & ASYNC_CLOSING) &&
		    (do_clocal || (uart_msr[MINOR(tty->device)] &
				   UART_MSR_DCD)))
			break;
		if (signal_pending(current)) {
			retval = -ERESTARTSYS;
			break;
		}
#ifdef SERIAL_DEBUG_OPEN
		printk("%s - blocking: bhnS%d, count = %d\n", __FUNCTION__,
						info->line, state->count);
#endif
		schedule();
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&info->open_wait, &wait);
	if (extra_count)
		state->count++;
	info->blocked_open--;
#ifdef SERIAL_DEBUG_OPEN
	printk("%s - after block: bhnS%d, count = %d\n", __FUNCTION__,
						info->line, state->count);
#endif
	if (retval)
		return retval;
	info->flags |= ASYNC_NORMAL_ACTIVE;
	return 0;
}

static int get_async_struct(int line, struct async_struct **ret_info)
{
	struct async_struct *info;
	struct serial_state *sstate;

	sstate = rs_table + line;
	sstate->count++;
	if (sstate->info) {
		*ret_info = sstate->info;
		return 0;
	}
	info = kmalloc(sizeof(struct async_struct), GFP_KERNEL);
	if (!info) {
		sstate->count--;
		return -ENOMEM;
	}
	memset(info, 0, sizeof(struct async_struct));
	init_waitqueue_head(&info->open_wait);
	init_waitqueue_head(&info->close_wait);
	init_waitqueue_head(&info->delta_msr_wait);
	info->magic = SERIAL_MAGIC;
	info->port = sstate->port;
	info->flags = sstate->flags;
	info->xmit_fifo_size = sstate->xmit_fifo_size;
	info->closing_wait = sstate->closing_wait;
	info->io_type = sstate->io_type;
	info->iomem_base = sstate->iomem_base;
	info->iomem_reg_shift = sstate->iomem_reg_shift;
	info->xmit_fifo_size = sstate->xmit_fifo_size = 128;
	info->line = line;
	info->state = sstate;
	if (sstate->info) {
		kfree(info);
		*ret_info = sstate->info;
		return 0;
	}
	*ret_info = sstate->info = info;
	return 0;
}

#ifdef USE_KDB_DEBUG
static void kprint_async(struct async_struct *i)
{
	kdb_printf("magic %d port %lx hub6 %d flags %x xmit fifo %d\n",
						i->magic, i->port, i->hub6,
						i->flags, i->xmit_fifo_size);
	kdb_printf("state %p tty %p rd mask %x ig mask %x timeout %d quot %d\n",
					i->state, i->tty, i->read_status_mask,
				i->ignore_status_mask, i->timeout, i->quot);
	kdb_printf("xchar %x cls delay %d cls wait %x cls wait 2 %x\n",
					i->x_char, i->close_delay,
					i->closing_wait, i->closing_wait2);
	kdb_printf("IER %x MCR %x LCR %x ACR %x\n", i->IER, i->MCR, i->LCR,
								i->ACR);
	kdb_printf("event %lx line %d blk open %d session %ld pgrp %ld\n",
					i->event, i->line, i->blocked_open,
							i->session, i->pgrp);
}

static void kprint_state(struct bhn_state *s)
{
	unsigned char *addr;
	struct cmd_state *cstate;

	addr = (unsigned char *)&s->cmd.addr.sin_addr;
	kdb_printf("state cmd  %d.%d.%d.%d:%d sock %p pid %d\n", addr[0],
				addr[1], addr[2], addr[3], s->cmd.addr.sin_port,
						s->cmd.sock, s->cmd.pid);

	addr = (unsigned char *)&s->data.addr.sin_addr;
	kdb_printf("state data %d.%d.%d.%d:%d sock %p pid %d\n", addr[0],
			addr[1], addr[2], addr[3], s->data.addr.sin_port,
						s->data.sock, s->data.pid);

	kdb_printf("state cb %p\n", s->event_cb);
	if (!list_empty(&s->cmdlist)) {
		kdb_printf("state cmds");
		list_for_each_entry(cstate, &s->cmdlist, list)
			kdb_printf(" %p %d", &cstate->cmd,
						cstate->cmd.pc_cmd.cmd);
		kdb_printf("\n");
	}
}

static void kprint_unit(struct bhn_unit *u)
{
	unsigned char *addr;
	int i;

	addr = (unsigned char *)&u->addr.sin_addr;
	kdb_printf("unit %d.%d.%d.%d:%d base %d baud %d ports\n", addr[0],
				addr[1], addr[2], addr[3], u->addr.sin_port,
						u->base_port, u->base_baud);

	for (i = 0; i < MAX_PORTS; i++) {
		if (u->ports[i])
			kdb_printf("%d:%p", i, u->ports[i]);
		if (i % 4)
			kdb_printf(" ");
		else
			kdb_printf("\n");
	}
	kdb_printf("\n");
}

static void kprint_map(struct bhn_map *m)
{
	kdb_printf("map unit %p port %d state %p\n", m->unit, m->unit_port,
								m->state);
}

static int bhn_kdb_async(int argc, const char **argv, const char **envp,
							struct pt_regs *regs)
{
	long minor;
	char *endp;
	struct serial_state *s;
	int i;

	if (argc > 2)
		return KDB_ARGCOUNT;

	if (argc == 1) {
		minor = simple_strtol(argv[1], &endp, 0);
		if (*endp)
			return KDB_BADINT;

		if (minor >= NPORTS) {
			kdb_printf("Invalid minor, must be between 0 and %d\n",
								NPORTS);
			return 0;
		}

		s = &rs_table[minor];
		if (s->info)
			kprint_async(s->info);
		else
			kdb_printf("Minor %ld has no async_struct\n", minor);
	} else
		for (i = 0, s = &rs_table[0]; i < NPORTS; i++, s++)
			if (s->info)
				kprint_async(s->info);

	return 0;
}

static int bhn_kdb_map(int argc, const char **argv, const char **envp,
							struct pt_regs *regs)
{
	long minor;
	char *endp;
	struct bhn_map *m;
	int i;

	if (argc > 2)
		return KDB_ARGCOUNT;

	if (argc == 1) {
		minor = simple_strtol(argv[1], &endp, 0);
		if (*endp)
			return KDB_BADINT;

		if (minor >= NPORTS) {
			kdb_printf("Invalid minor, must be between 0 and %d\n",
								NPORTS);
			return 0;
		}

		m = &bhn_table[minor];
		if (m->unit) {
			kprint_map(m);
			kprint_unit(m->unit);
			if (m->state)
				kprint_state(m->state);
		} else
			kdb_printf("Minor %ld is not mapped\n", minor);
	} else
		for (i = 0, m = &bhn_table[0]; i < NPORTS; i++, m++)
			if (m->unit) {
				kprint_map(m);
				kprint_unit(m->unit);
				if (m->state)
					kprint_state(m->state);
			}

	return 0;
}

static int bhn_kdb_unit(int argc, const char **argv, const char **envp,
							struct pt_regs *regs)
{
	struct bhn_unit *u;
	int i;
	int j;

	for (i = 0, u = &bhn_unit_table[0]; i < NPORTS; i++, u++)
		for (j = 0; j < MAX_PORTS; j++)
			if (u->ports[j]) {
				kprint_unit(u);
				break;
			}

	return 0;
}
#endif

int __init bhn_init(void)
{
	struct serial_state *state;
	int err;
	int i;

	if (ASYNC_INTERNAL_FLAGS & ASYNC_QUERY_IN_PROGRESS) {
		printk("%s: Warning: The QUERY_IN_PROGRESS flag has been usurped.\n", __FUNCTION__);
		printk("%s: Warning: Please contact support@connecttech.com\n", __FUNCTION__);
	}

	memset(&bhn_driver, 0, sizeof(struct tty_driver));
	bhn_driver.magic = TTY_DRIVER_MAGIC;
#if (LINUX_VERSION_CODE > 0x20100)
	bhn_driver.driver_name = bhn_drvname;
#endif
#if (LINUX_VERSION_CODE > 0x2032D && defined(CONFIG_DEVFS_FS))
	bhn_driver.name = "bhn/%d";
#else
	bhn_driver.name = "bhnS";
#endif
	bhn_driver.num = NPORTS;
	bhn_driver.type = TTY_DRIVER_TYPE_SERIAL;
	bhn_driver.subtype = SERIAL_TYPE_NORMAL;
	bhn_driver.init_termios = tty_std_termios;
	bhn_driver.init_termios.c_cflag =
	    B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	bhn_driver.flags = TTY_DRIVER_REAL_RAW;
	bhn_driver.refcount = &bhn_refcount;
	bhn_driver.table = bhn_tty;
	bhn_driver.termios = bhn_termios;
	bhn_driver.termios_locked = bhn_termios_locked;

	bhn_driver.open = bhn_open;
	bhn_driver.close = bhn_close;
	bhn_driver.write = bhn_write;
	bhn_driver.flush_chars = bhn_flush_chars;
	bhn_driver.write_room = bhn_write_room;
	bhn_driver.chars_in_buffer = bhn_chars_in_buffer;
	bhn_driver.ioctl = bhn_ioctl;
	bhn_driver.set_termios = bhn_set_termios;
	bhn_driver.throttle = bhn_throttle;
	bhn_driver.unthrottle = bhn_unthrottle;
	bhn_driver.stop = bhn_stop;
	bhn_driver.start = bhn_start;
	bhn_driver.hangup = bhn_hangup;
#if (LINUX_VERSION_CODE >= 0x020142)
	bhn_driver.break_ctl = bhn_break;
#endif
	bhn_driver.flush_buffer = bhn_flush_buffer;
#if (LINUX_VERSION_CODE >= 0x02010f)
	bhn_driver.wait_until_sent = bhn_wait_until_sent;
	bhn_driver.send_xchar = bhn_send_xchar;
#endif
#ifdef CONFIG_PROC_FS
	bhn_driver.read_proc = bhn_read_proc;
	bhn_driver.write_proc = bhn_write_proc;
#endif

	bhn_callout_driver = bhn_driver;
#if (LINUX_VERSION_CODE > 0x2032D && defined(CONFIG_DEVFS_FS))
	bhn_callout_driver.name = "bhnc/%d";
#else
	bhn_callout_driver.name = "bhnC";
#endif
	bhn_callout_driver.subtype = SERIAL_TYPE_CALLOUT;
#if (LINUX_VERSION_CODE >= 0x02010f)
	bhn_callout_driver.read_proc = 0;
	bhn_callout_driver.write_proc = 0;
#endif

	err = tty_register_driver(&bhn_driver);
	if (err < 0) {
		printk("Couldn't register %s driver\n", bhn_drvname);
		goto out;
	}
	err = tty_register_driver(&bhn_callout_driver);
	if (err < 0) {
		printk("Couldn't register %s callout driver\n", bhn_drvname);
		goto unreg1;
	}

	memset(&rs_table, 0, sizeof(rs_table));
	for (i = 0, state = rs_table; i < NPORTS; i++,state++) {
		state->magic = SSTATE_MAGIC;
		state->line = i;
		state->type = PORT_UNKNOWN;
		state->xmit_fifo_size = 128;
		state->custom_divisor = 0;
		state->close_delay = 5*HZ/10;
		state->closing_wait = 30*HZ;
		state->icount.cts = state->icount.dsr =
			state->icount.rng = state->icount.dcd = 0;
		state->icount.rx = state->icount.tx = 0;
		state->icount.frame = state->icount.parity = 0;
		state->icount.overrun = state->icount.brk = 0;
	}

#ifdef CONFIG_PROC_FS
	err = -ENOMEM;
	bhn_proc_root = proc_mkdir(PROC_BHNROOT_NM, 0);
	if (!bhn_proc_root) {
		printk("Couldn't set up proc fs '%s'\n", PROC_BHNROOT_NM);
		goto unreg2;
	}

	bhn_proc_autodiscover = create_proc_entry(PROC_AUTOD_NM, 0444,
								bhn_proc_root);
	if (!bhn_proc_autodiscover) {
		printk("Couldn't set up proc fs '%s'\n", PROC_AUTOD_NM);
		goto unproc1;
	}
	bhn_proc_autodiscover->read_proc = bhn_proc_autod_read;
	bhn_proc_autodiscover->write_proc = bhn_proc_autod_write;

	bhn_proc_config = create_proc_entry(PROC_CONFIG_NM, 0644,
								bhn_proc_root);
	if (!bhn_proc_config) {
		printk("Couldn't set up proc fs '%s'\n", PROC_CONFIG_NM);
		goto unproc2;
	}
	bhn_proc_config->read_proc = bhn_proc_config_read;
	bhn_proc_config->write_proc = bhn_proc_config_write;

	bhn_proc_stats = create_proc_entry(PROC_STATS_NM, 0444, bhn_proc_root);
	if (!bhn_proc_stats) {
		printk("Couldn't set up proc fs '%s'\n", PROC_STATS_NM);
		goto unproc3;
	}
	bhn_proc_stats->read_proc = bhn_read_proc;
#endif

	memset(&bhn_unit_table, 0, sizeof(bhn_unit_table));
	memset(&bhn_table, 0, sizeof(bhn_table));

	for (i = 0; i < NPORTS; i++) {
		struct bhn_map *map = &bhn_table[i];

		uart_lsr[i] = UART_LSR_TEMT;
		sema_init(&map->openlock, 1);
		spin_lock_init(&map->fliplock);
		sema_init(&map->auto_reconn_lock, 1);
		init_completion(&map->auto_reconn_comp);
		sema_init(&map->cr_lock, 1);
		sema_init(&map->cw_lock, 1);
		sema_init(&map->dr_lock, 1);
		sema_init(&map->dw_lock, 1);
	}

#ifdef USE_KDB_DEBUG
	kdb_register("async", bhn_kdb_async, "[minor]",
						"Displays async_struct", 0);
	kdb_register("map", bhn_kdb_map, "[minor]",
						"Displays bhn map table", 0);
	kdb_register("unit", bhn_kdb_unit, "", "Displays bhn unit table", 0);
#endif

	printk(KERN_INFO "%s driver version %s (%s)\n", bhn_drvname,
	       bhn_version, bhn_revdate);

	err = 0;
	goto out;

#ifdef CONFIG_PROC_FS
unproc3:
	remove_proc_entry(PROC_CONFIG_NM, bhn_proc_root);
unproc2:
	remove_proc_entry(PROC_AUTOD_NM, bhn_proc_root);
unproc1:
	remove_proc_entry(PROC_BHNROOT_NM, NULL);
#endif
unreg2:
	tty_unregister_driver(&bhn_callout_driver);
unreg1:
	tty_unregister_driver(&bhn_driver);
out:
	return err;
}

static void __exit bhn_fini(void)
{
	int err;
#ifdef USE_KDB_DEBUG
	kdb_unregister("async");
	kdb_unregister("map");
	kdb_unregister("unit");
#endif
#ifdef CONFIG_PROC_FS
	remove_proc_entry(PROC_STATS_NM, bhn_proc_root);
	remove_proc_entry(PROC_CONFIG_NM, bhn_proc_root);
	remove_proc_entry(PROC_AUTOD_NM, bhn_proc_root);
	remove_proc_entry(PROC_BHNROOT_NM, NULL);
#endif

	err = tty_unregister_driver(&bhn_driver);
	if (err)
		printk("serial: failed to unregister %s driver (%d)\n",
		       bhn_drvname, err);
	err = tty_unregister_driver(&bhn_callout_driver);
	if (err)
		printk
		    ("serial: failed to unregister %s callout driver (%d)\n",
		     bhn_drvname, err);
}

module_init(bhn_init);
module_exit(bhn_fini);
MODULE_DESCRIPTION("Blue Heat/Net virtual serial port driver");
MODULE_AUTHOR("Stuart MacDonald <stuartm@connecttech.com>");
MODULE_LICENSE("GPL");

static void startup(struct async_struct *info)
{
	int inited;
	unsigned long flags;

	save_flags(flags); cli();
	inited = info->flags & ASYNC_INITIALIZED;
	restore_flags(flags);

	if (inited)
		return;

#ifdef SERIAL_DEBUG_OPEN
	printk("bhn: %s bhnS%d...", __FUNCTION__, info->line);
#endif

	info->MCR = 0;
	if (info->tty->termios->c_cflag & CBAUD)
		info->MCR = UART_MCR_DTR | UART_MCR_RTS;
	prot_set_rts_dtr(MINOR(info->tty->device), info->MCR & UART_MCR_RTS,
						info->MCR & UART_MCR_DTR);

	if (info->tty)
		clear_bit(TTY_IO_ERROR, &info->tty->flags);

	/*
	 * Set up the tty->alt_speed kludge
	 */
#if (LINUX_VERSION_CODE >= 131394)	/* Linux 2.1.66 */
	if (info->tty) {
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
			info->tty->alt_speed = 57600;
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
			info->tty->alt_speed = 115200;
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
			info->tty->alt_speed = 230400;
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
			info->tty->alt_speed = 460800;
	}
#endif

	/* set the speed of the serial port */
	change_speed(info, 0);

	save_flags(flags); cli();
	info->flags |= ASYNC_INITIALIZED;
	restore_flags(flags);
#ifdef SERIAL_DEBUG_OPEN
	printk("bhn: %s bhnS%d done\n", __FUNCTION__, info->line);
#endif
}

/*
 * This routine will shutdown a serial port; interrupts are disabled, and
 * DTR is dropped if the hangup on close termio flag is on.
 */
static void shutdown(struct async_struct *info)
{
    int minor = -1;
	struct serial_state *state;

    if(info->tty)
        minor = MINOR(info->tty->device);

	if (!(info->flags & ASYNC_INITIALIZED))
		return;

	state = info->state;

#ifdef SERIAL_DEBUG_OPEN
	printk("Shutting down serial port %d (irq %d)....", info->line,
	       state->irq);
#endif

	/*
	 * clear delta_msr_wait queue to avoid mem leaks: we may free the irq
	 * here so the queue might never be waken up
	 */
	wake_up_interruptible(&info->delta_msr_wait);

	info->IER = 0;
    if(minor != -1)
        prot_set_event(minor, 0);  /* disable all events */

	if (!info->tty || (info->tty->termios->c_cflag & HUPCL))
		info->MCR &= ~(UART_MCR_DTR | UART_MCR_RTS);
    if(minor != -1)
        prot_set_rts_dtr(minor, info->MCR & UART_MCR_RTS,
                         info->MCR & UART_MCR_DTR);

	/* disable FIFO's */
	/* The protocol doesn't have a way to disable the FIFO */
    /*
    if(minor != -1)
        prot_fifo_control(minor, 0, 0);
    */
	if (info->tty)
		set_bit(TTY_IO_ERROR, &info->tty->flags);

	info->flags &= ~ASYNC_INITIALIZED;
}

void bhn_do_change_speed(int minor)
{
	struct async_struct *info = rs_table[minor].info;

	change_speed(info, 0);
}

static void change_speed(struct async_struct *info, struct termios *old)
{
	int minor = MINOR(info->tty->device);
	int quot = 0, baud_base, baud;
	unsigned cflag, fcr = 0;
	int bits;
	int event;

#ifdef SERIAL_DEBUG_IOCTL
	printk("change_speed: %d\n", minor);
#endif
	if (!info->tty || !info->tty->termios) {
		printk("change_speed: info->tty %p info->tty->termios %p\n",
						info->tty, info->tty->termios);
		return;
	}
	cflag = info->tty->termios->c_cflag;

	switch (cflag & CSIZE) {
		default:
		case CS5: bits = 7; break;
		case CS6: bits = 8; break;
		case CS7: bits = 9; break;
		case CS8: bits = 10; break;
	}
	if (cflag & CSTOPB)
		bits++;
	if (cflag & PARENB)
		bits++;

	/* Determine divisor based on baud rate */
	baud = tty_get_baud_rate(info->tty);
	if(!baud)
		baud = 9600;
	baud_base = info->state->baud_base = bhn_table[minor].unit->base_baud;
	if (baud == 38400 && ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_CUST))
		quot = info->state->custom_divisor;
	else {
		if (baud == 134)
			/* Special case since 134 is really 134.5 */
			quot = (2*baud_base / 269);
		else if (baud)
			quot = baud_base / baud;
	}
	/* If the quotient is zero refuse the change */
	if (!quot && old) {
		info->tty->termios->c_cflag &= ~CBAUD;
		info->tty->termios->c_cflag |= (old->c_cflag & CBAUD);
		baud = tty_get_baud_rate(info->tty);
		if (!baud)
			baud = 9600;
		if (baud == 38400 &&
		    ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_CUST))
			quot = info->state->custom_divisor;
		else {
			if (baud == 134)
				/* Special case since 134 is really 134.5 */
				quot = (2*baud_base / 269);
			else if (baud)
				quot = baud_base / baud;
		}
	}
	/* As a last resort, if the quotient is zero, default to 9600 bps */
	if (!quot)
		quot = baud_base / 9600;
	info->quot = quot;
	/*
	 * Add 4096 to account for possible data still in the TX
	 * buffers on the BHN.  If this size ever changes the driver
	 * needs to be updated.
	 * Should also account for the network buffers.
	 */
	baud = baud_base / quot;
	info->timeout = (4096 + info->xmit_fifo_size) * bits * HZ;
	info->timeout /= baud;

	/* Set up FIFO's */
	if ((info->state->baud_base / quot) < 2400)
		fcr = 1;
	else
		fcr = 2;

	/* CTS flow control flag and modem status interrupts */
	info->IER &= ~UART_IER_MSI;
	if (info->flags & ASYNC_HARDPPS_CD)
		info->IER |= UART_IER_MSI;
	if (cflag & CRTSCTS) {
		info->flags |= ASYNC_CTS_FLOW;
		info->IER |= UART_IER_MSI;
	} else
		info->flags &= ~ASYNC_CTS_FLOW;
	if (cflag & CLOCAL)
		info->flags &= ~ASYNC_CHECK_CD;
	else {
		info->flags |= ASYNC_CHECK_CD;
		info->IER |= UART_IER_MSI;
	}

	event = 0;
	if(info->IER & UART_IER_MSI)
		event |= EVENT_CTS | EVENT_DSR | EVENT_RING | EVENT_DCD;
	if(info->IER & UART_IER_RLSI)
		event |= EVENT_OE | EVENT_PE | EVENT_FE;
	if(info->IER & UART_IER_THRI)
		event |= EVENT_TXFIN;

	/* for now, we just turn on all events except the TXFIN */
	event = EVENT_ALL & (~EVENT_TXFIN);

	/*
	 * Set up parity check flag
	 */
#define RELEVANT_IFLAG(iflag) (iflag & (IGNBRK|BRKINT|IGNPAR|PARMRK|INPCK))

	info->read_status_mask = UART_LSR_OE | UART_LSR_THRE | UART_LSR_DR;
	if (I_INPCK(info->tty))
		info->read_status_mask |= UART_LSR_FE | UART_LSR_PE;
	if (I_BRKINT(info->tty) || I_PARMRK(info->tty))
		info->read_status_mask |= UART_LSR_BI;

	/*
	 * Characters to ignore
	 */
	info->ignore_status_mask = 0;
	if (I_IGNPAR(info->tty))
		info->ignore_status_mask |= UART_LSR_PE | UART_LSR_FE;
	if (I_IGNBRK(info->tty)) {
		info->ignore_status_mask |= UART_LSR_BI;
		/*
		 * If we're ignore parity and break indicators, ignore
		 * overruns too.  (For real raw support).
		 */
		if (I_IGNPAR(info->tty))
			info->ignore_status_mask |= UART_LSR_OE;
	}
	/*
	 * !!! ignore all characters if CREAD is not set
	 */
	if ((cflag & CREAD) == 0)
		info->ignore_status_mask |= UART_LSR_DR;

	/* set to BlueHeat/Net */
	prot_set_line(minor, info->tty);
	/* prot_fifo_control(minor, fcr, fcr); */
	prot_set_event(minor, EVENT_ALL);
	prot_get_msr(minor, &uart_msr[minor]);

	return;
}

static int get_serial_info(struct async_struct * info,
			   struct serial_struct * retinfo)
{
	struct serial_struct tmp;
	struct serial_state *state = info->state;

	if (!retinfo)
		return -EFAULT;
	memset(&tmp, 0, sizeof(tmp));
	tmp.type = state->type;
	tmp.line = state->line;
	tmp.port = state->port;
	tmp.port_high = 0;
	tmp.irq = state->irq;
	tmp.flags = state->flags;
	tmp.xmit_fifo_size = state->xmit_fifo_size;
	tmp.baud_base = state->baud_base;
	tmp.close_delay = state->close_delay;
	tmp.closing_wait = state->closing_wait;
	tmp.custom_divisor = state->custom_divisor;
	tmp.hub6 = state->hub6;
	tmp.io_type = state->io_type;
	if (copy_to_user(retinfo,&tmp,sizeof(*retinfo)))
		return -EFAULT;
	return 0;
}

static int set_serial_info(struct async_struct * info,
			   struct serial_struct * new_info)
{
	struct serial_struct new_serial;
 	struct serial_state old_state, *state;
	unsigned int		i,change_irq,change_port;
	unsigned long		new_port;

	if (copy_from_user(&new_serial,new_info,sizeof(new_serial)))
		return -EFAULT;
	state = info->state;
	old_state = *state;

	new_port = new_serial.port;
	change_irq = 0;  /* doesn't apply to BHN */
	change_port = 0;

	if (!capable(CAP_SYS_ADMIN)) {
		if (change_irq || change_port ||
		    (new_serial.baud_base != state->baud_base) ||
		    (new_serial.type != state->type) ||
		    (new_serial.close_delay != state->close_delay) ||
		    (new_serial.xmit_fifo_size != state->xmit_fifo_size) ||
		    ((new_serial.flags & ~ASYNC_USR_MASK) !=
		     (state->flags & ~ASYNC_USR_MASK)))
			return -EPERM;
		state->flags = ((state->flags & ~ASYNC_USR_MASK) |
			       (new_serial.flags & ASYNC_USR_MASK));
		info->flags = ((info->flags & ~ASYNC_USR_MASK) |
			       (new_serial.flags & ASYNC_USR_MASK));
		state->custom_divisor = new_serial.custom_divisor;
		goto check_and_exit;
	}

	new_serial.irq = irq_cannonicalize(new_serial.irq);

	if ((new_serial.irq >= NR_IRQS) || (new_serial.irq < 0) ||
	    (new_serial.baud_base < 9600)|| (new_serial.type < PORT_UNKNOWN) ||
	    (new_serial.type > PORT_MAX) || (new_serial.type == PORT_CIRRUS) ||
	    (new_serial.type == PORT_STARTECH)) {
		return -EINVAL;
	}

	if ((new_serial.type != state->type) ||
	    (new_serial.xmit_fifo_size <= 0))
		new_serial.xmit_fifo_size = 64;

	/* Make sure address is not already in use */
	if (new_serial.type) {
		for (i = 0 ; i < NPORTS; i++)
			if ((state != &rs_table[i]) &&
			    (rs_table[i].io_type == SERIAL_IO_PORT) &&
			    (rs_table[i].port == new_port) &&
			    rs_table[i].type)
				return -EADDRINUSE;
	}

	if ((change_port || change_irq) && (state->count > 1))
		return -EBUSY;

	/*
	 * OK, past this point, all the error checking has been done.
	 * At this point, we start making changes.....
	 */

	state->baud_base = new_serial.baud_base;
	state->flags = ((state->flags & ~ASYNC_FLAGS) |
			(new_serial.flags & ASYNC_FLAGS));
	info->flags = ((state->flags & ~ASYNC_INTERNAL_FLAGS) |
		       (info->flags & ASYNC_INTERNAL_FLAGS));
	state->custom_divisor = new_serial.custom_divisor;
	state->close_delay = new_serial.close_delay * HZ/100;
	state->closing_wait = new_serial.closing_wait * HZ/100;
#if (LINUX_VERSION_CODE > 0x20100)
	info->tty->low_latency = 1; /*(info->flags & ASYNC_LOW_LATENCY) ? 1 : 0;*/
#endif
	info->xmit_fifo_size = state->xmit_fifo_size =
		new_serial.xmit_fifo_size;
/*
	if ((state->type != PORT_UNKNOWN) && state->port) {
		release_region(state->port,8);
	}
*/
	state->type = new_serial.type;
	if (change_port || change_irq) {
		/*
		 * We need to shutdown the serial port at the old
		 * port/irq combination.
		 */
		shutdown(info);
		state->irq = new_serial.irq;
		info->port = state->port = new_port;
		info->hub6 = state->hub6 = new_serial.hub6;
		if (info->hub6)
			info->io_type = state->io_type = SERIAL_IO_HUB6;
		else if (info->io_type == SERIAL_IO_HUB6)
			info->io_type = state->io_type = SERIAL_IO_PORT;
	}
/*
	if ((state->type != PORT_UNKNOWN) && state->port) {
			request_region(state->port,8,"serial(set)");
	}
*/

check_and_exit:
	if ((!state->port && !state->iomem_base) || !state->type)
		return 0;
	if (info->flags & ASYNC_INITIALIZED) {
		if (((old_state.flags & ASYNC_SPD_MASK) !=
		     (state->flags & ASYNC_SPD_MASK)) ||
		    (old_state.custom_divisor != state->custom_divisor)) {
#if (LINUX_VERSION_CODE >= 131394) /* Linux 2.1.66 */
			if ((state->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
				info->tty->alt_speed = 57600;
			if ((state->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
				info->tty->alt_speed = 115200;
			if ((state->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
				info->tty->alt_speed = 230400;
			if ((state->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
				info->tty->alt_speed = 460800;
#endif
			change_speed(info, 0);
		}
	} else
		startup(info);
	return 0;
}

/*
 * get_lsr_info - get line status register info
 *
 * Purpose: Let user call ioctl() to get info when the UART physically
 * 	    is emptied.  On bus types like RS485, the transmitter must
 * 	    release the bus after transmitting. This must be done when
 * 	    the transmit shift register is empty, not be done when the
 * 	    transmit holding register is empty.  This functionality
 * 	    allows an RS485 driver to be written in user space.
 */
static int get_lsr_info(struct async_struct *info, unsigned int *value)
{
	int minor = MINOR(info->tty->device);
	unsigned int result;

	result = ((uart_lsr[minor] & UART_LSR_TEMT) ? TIOCSER_TEMT : 0);

	/*
	 * If we're about to load something into the transmit
	 * register, we'll pretend the transmitter isn't empty to
	 * avoid a race condition (depending on when the transmit
	 * interrupt happens).
	 */
	if (info->x_char)
		result &= ~TIOCSER_TEMT;

	if (copy_to_user(value, &result, sizeof(unsigned int)))
		return -EFAULT;

	return 0;
}

static int get_modem_info(struct async_struct *info, unsigned int *value)
{
	int minor = MINOR(info->tty->device);
	unsigned int result;

	result =  ((info->MCR & UART_MCR_RTS) ? TIOCM_RTS : 0)
		| ((info->MCR & UART_MCR_DTR) ? TIOCM_DTR : 0)
		| ((uart_msr[minor] & UART_MSR_DCD) ? TIOCM_CAR : 0)
		| ((uart_msr[minor] & UART_MSR_RI) ? TIOCM_RNG : 0)
		| ((uart_msr[minor] & UART_MSR_DSR) ? TIOCM_DSR : 0)
		| ((uart_msr[minor] & UART_MSR_CTS) ? TIOCM_CTS : 0);

	if (copy_to_user(value, &result, sizeof(int)))
		return -EFAULT;

	return 0;
}

static int set_modem_info(struct async_struct * info, unsigned int cmd,
			  unsigned int *value)
{
    int minor = MINOR(info->tty->device);
	unsigned int arg;

	if (copy_from_user(&arg, value, sizeof(int)))
		return -EFAULT;

	switch (cmd) {
	case TIOCMBIS:
		if (arg & TIOCM_RTS)
			info->MCR |= UART_MCR_RTS;
		if (arg & TIOCM_DTR)
			info->MCR |= UART_MCR_DTR;
		if (arg & TIOCM_LOOP)
			info->MCR |= UART_MCR_LOOP;
		break;
	case TIOCMBIC:
		if (arg & TIOCM_RTS)
			info->MCR &= ~UART_MCR_RTS;
		if (arg & TIOCM_DTR)
			info->MCR &= ~UART_MCR_DTR;
		if (arg & TIOCM_LOOP)
			info->MCR &= ~UART_MCR_LOOP;
		break;
	case TIOCMSET:
		info->MCR = ((info->MCR & ~(UART_MCR_RTS |
					    UART_MCR_DTR |
					    UART_MCR_LOOP))
			     | ((arg & TIOCM_RTS) ? UART_MCR_RTS : 0)
			     | ((arg & TIOCM_DTR) ? UART_MCR_DTR : 0)
			     | ((arg & TIOCM_LOOP) ? UART_MCR_LOOP : 0));
		break;
	default:
		return -EINVAL;
	}

    /* we have no way to set other fields */
    prot_set_rts_dtr(minor, info->MCR & UART_MCR_RTS, info->MCR & UART_MCR_DTR);

	return 0;
}


static int bhn_network_up(int minor) {
	int err;

	err = prot_tcp_start(minor, &event_handler);
	if (err < 0) {
		if (err != -ERESTARTSYS)
			printk("bhn: Couldn't start network connection: %s (%d)\n",
				(elist[-err] ? elist[-err] : "<no text>"), err);
		goto out;
	}

	err = prot_open(minor);
	if (err < 0) {
		if (err != -ERESTARTSYS)
			printk("bhn: Couldn't open remote port: %s (%d)\n",
				(elist[-err] ? elist[-err] : "<no text>"), err);
		prot_tcp_stop(minor);
	}

out:
	return err;
}

static int bhn_open(struct tty_struct *tty, struct file *filp)
{
	int minor = MINOR(tty->device);
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_unit *unit = map->unit;
	struct async_struct *info;
	int err;

	MOD_INC_USE_COUNT;
#ifdef SERIAL_DEBUG_OPEN
	printk("%s - bhnS%d by pid %d\n", __FUNCTION__, minor, current->pid);
#endif

	if (!unit) {
		err = -ENODEV;
		MOD_DEC_USE_COUNT;
		goto out;
	}

	if (down_interruptible(&map->openlock)) {
		printk("%s - interrupted while acquiring open lock\n",
								__FUNCTION__);
		err = -ERESTARTSYS;
		MOD_DEC_USE_COUNT;
		goto out;
	}
	BHNDEBUG_OPENLOCK(("%s - pid=%d down open lock 0x%p\n", __FUNCTION__,
						current->pid, &map->openlock));
	/* get_async_struct() increments state->count */
	err = get_async_struct(minor, &info);
	if (err) {
		MOD_DEC_USE_COUNT;
		goto lock_out;
	}

	tty->driver_data = info;
	info->tty = tty;
        info->tty->low_latency = 1;

	/* if the port is the middle of closing, bail out now */
	if (tty_hung_up_p(filp) || (info->flags & ASYNC_CLOSING)) {
		BHNDEBUG_OPENLOCK(("%s - pid=%d up open lock 0x%p\n",
				__FUNCTION__, current->pid, &map->openlock));
		up(&map->openlock);
		if (info->flags & ASYNC_CLOSING)
			interruptible_sleep_on(&info->close_wait);
#ifdef SERIAL_DO_RESTART
		return ((info->flags & ASYNC_HUP_NOTIFY) ?  -EAGAIN :
								-ERESTARTSYS);
#else
		return -EAGAIN;
#endif
	}

	down(&map->auto_reconn_lock);
	map->flags |= MAP_OPENING;
	up(&map->auto_reconn_lock);

	/* first open? */
	if (info->state->count == 1) {
		err = bhn_network_up(minor);
		if (err < 0)
			goto flag_out;
	}

	/* we have a connection to the BHN, set up the async stuff */
	startup(info);

	err = block_til_ready(tty, filp, info);
	if (err)
		goto flag_out;

	if ((info->state->count == 1) &&
	    (info->flags & ASYNC_SPLIT_TERMIOS)) {
		if (tty->driver.subtype == SERIAL_TYPE_NORMAL)
			*tty->termios = info->state->normal_termios;
		else
			*tty->termios = info->state->callout_termios;
		change_speed(info, 0);
	}
	info->session = current->session;
	info->pgrp = current->pgrp;
#ifdef SERIAL_DEBUG_OPEN
	printk("%s - success\n", __FUNCTION__);
#endif

flag_out:
	down(&map->auto_reconn_lock);
	map->flags &= ~MAP_OPENING;
	up(&map->auto_reconn_lock);

lock_out:
	BHNDEBUG_OPENLOCK(("%s pid=%d up open lock 0x%p\n", __FUNCTION__,
						current->pid, &map->openlock));
	up(&map->openlock);

out:
#ifdef SERIAL_DEBUG_OPEN
	printk("%s - returns %d\n", err);
#endif
	return err;
}

/*
 * ------------------------------------------------------------
 * bhn_close()
 *
 * This routine is called when the serial port gets closed.  First, we
 * wait for the last remaining data to be sent.
 * ------------------------------------------------------------
 */

static int bhn_network_down(int minor) {
	int err;
	int ret = 0;

	err = prot_close(minor);
	if (err < 0)
		ret = err;
	err = prot_tcp_stop(minor);
	if (!ret && err < 0)
		ret = err;
#ifdef SERIAL_DEBUG_EVENTS
	printk("%s - bhnS%d setting UART_LSR_TEMT\n", __FUNCTION__, minor);
#endif
	uart_lsr[minor] |= UART_LSR_TEMT;

	return ret;
}

static void bhn_close(struct tty_struct *tty, struct file *filp)
{
	struct async_struct *info = tty->driver_data;
	struct serial_state *state;
	int minor = MINOR(tty->device);
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *bstate = map->state;

#ifdef SERIAL_DEBUG_OPEN
	printk("%s - bhnS%d\n", __FUNCTION__, minor);
#endif

	/*
	 * If the port is not configured, bhn_open returns -ENODEV, the TTY
	 * layer still calls bhn_close. Guard against that.
	 */
	if (!info)
		return;
	if (!info->state)
		return;
	state = info->state;

	if (tty_hung_up_p(filp)) {
		DBG_CNT("before DEC-hung");
		MOD_DEC_USE_COUNT;
		return;
	}

	if (down_interruptible(&map->openlock)) {
		printk("%s interrupted while acquiring open lock\n",
								__FUNCTION__);
		MOD_DEC_USE_COUNT;
		return;
	}
	BHNDEBUG_OPENLOCK(("%s - pid=%d down open lock 0x%p\n", __FUNCTION__,
						current->pid, &map->openlock));

#ifdef SERIAL_DEBUG_OPEN
	printk("%s - bhnS%d, count = %d\n", __FUNCTION__, info->line,
								state->count);
#endif

	if ((tty->count == 1) && (state->count != 1)) {
		/*
		 * Uh, oh.  tty->count is 1, which means that the tty
		 * structure will be freed.  state->count should always
		 * be one in these conditions.  If it's greater than
		 * one, we've got real problems, since it means the
		 * serial port won't be shutdown.
		 */
		printk("%s - bad serial port count; tty->count is 1, "
			"state->count is %d\n", __FUNCTION__, state->count);
		state->count = 1;
	}
	if (--state->count < 0) {
		printk("%s - bad serial port count for bhnS%d: %d\n",
				__FUNCTION__, info->line, state->count);
		state->count = 0;
	}
	if (state->count) {
		DBG_CNT("before DEC-2");
		BHNDEBUG_OPENLOCK(("%s pid=%d up open lock 0x%p\n",
				__FUNCTION__, current->pid, &map->openlock));
		up(&map->openlock);
		MOD_DEC_USE_COUNT;
		return;
	}
	info->flags |= ASYNC_CLOSING;
	down(&map->auto_reconn_lock);
	map->flags |= MAP_CLOSING;
	up(&map->auto_reconn_lock);

	/*
	 * Save the termios structure, since this port may have
	 * separate termios for callout and dialin.
	 */
	if (info->flags & ASYNC_NORMAL_ACTIVE)
		info->state->normal_termios = *tty->termios;
	if (info->flags & ASYNC_CALLOUT_ACTIVE)
		info->state->callout_termios = *tty->termios;

	if (bstate && (map->flags & MAP_OPENED)) {
		/*
	 	* Now we wait for the transmit buffer to clear; and we notify
	 	* the line discipline to only process XON/XOFF characters.
	 	*/
		tty->closing = 1;
		if (info->closing_wait != ASYNC_CLOSING_WAIT_NONE)
			tty_wait_until_sent(tty, info->closing_wait);

		if (info->flags & ASYNC_INITIALIZED)
			bhn_wait_until_sent(tty, info->timeout);
		shutdown(info);
		if (tty->driver.flush_buffer)
			tty->driver.flush_buffer(tty);
		if (tty->ldisc.flush_buffer)
			tty->ldisc.flush_buffer(tty);
		tty->closing = 0;
	}

	/* last close? */
	if (!state->count) {
		down(&map->auto_reconn_lock);
		if (map->auto_reconn_pid) {
			up(&map->auto_reconn_lock);
			kill_proc(map->auto_reconn_pid, SIGKILL, 1);
			wait_for_completion(&map->auto_reconn_comp);
		} else
			up(&map->auto_reconn_lock);
		bhn_network_down(minor);
	}

	BHNDEBUG_OPENLOCK(("%s pid=%d up open lock 0x%p\n", __FUNCTION__,
						current->pid, &map->openlock));
	up(&map->openlock);

	info->event = 0;
	info->tty = 0;
	if (info->blocked_open) {
		if (info->close_delay) {
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(info->close_delay);
		}
		wake_up_interruptible(&info->open_wait);
	}
	info->flags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CALLOUT_ACTIVE |
								ASYNC_CLOSING);
	down(&map->auto_reconn_lock);
	map->flags &= ~MAP_CLOSING;
	up(&map->auto_reconn_lock);
	wake_up_interruptible(&info->close_wait);
	MOD_DEC_USE_COUNT;
#ifdef SERIAL_DEBUG_OPEN
	printk("%s - bhnS%d closed\n", __FUNCTION__, info->line);
#endif
}

int bhn_auto_reconnect(void *arg)
{
	int minor = (int)arg;
	struct bhn_map *map = &bhn_table[minor];
	int err;

	daemonize();
	snprintf(current->comm, sizeof(current->comm), "bhnS%d ar", minor);
	current->tty = NULL;
	spin_lock_irq(&current->sigmask_lock);
	siginitsetinv(&current->blocked, sigmask(SIGKILL));
	flush_signals(current);
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);
	reparent_to_init();

	complete(&map->auto_reconn_comp);

	bhn_network_down(minor);
	do {
		err = bhn_network_up(minor);
		if (err == -ERESTARTSYS)
			break;
	} while (err);

	map->auto_reconn_pid = 0;
	complete_and_exit(&map->auto_reconn_comp, 0);
}

static void event_handler(int minor, struct bhn_event_occurred *e)
{
	struct serial_state *state;
	struct async_struct *info;
	struct tty_struct *tty;
	struct bhn_map *map;
	unsigned int events = htons(e->events);
	unsigned char status;
	unsigned long flags;
	int push = 0;

	state = rs_table + minor;
	if (!state->info) {
		/* this port is probably not even open */
		printk("bhn: Event %x line %d but no info\n", events, minor);
		return;
	}
	info = state->info;
	tty = info->tty;

	/* update the MSR if any of the events affect MSR */
	if (events & (EVENT_DCD | EVENT_RING | EVENT_DSR | EVENT_CTS)) {
		uart_msr[minor] = e->msr;
		wake_up_interruptible(&info->delta_msr_wait);
	}
	status = uart_msr[minor];

	map = bhn_table + minor;
	spin_lock_irqsave(&map->fliplock, flags);
	if (events & EVENT_OE) {
		state->icount.overrun += htonl(e->oe_count);
		if (info->read_status_mask & UART_LSR_OE) {
			++push;
			*tty->flip.char_buf_ptr = 0;
			*tty->flip.flag_buf_ptr = TTY_OVERRUN;
			tty->flip.char_buf_ptr++;
			tty->flip.flag_buf_ptr++;
			tty->flip.count++;
		}
		events &= ~EVENT_OE;
	}
	if (events & EVENT_PE) {
		state->icount.parity += htonl(e->pe_count);
		if (info->read_status_mask & UART_LSR_PE) {
			++push;
			*tty->flip.char_buf_ptr = 0;
			*tty->flip.flag_buf_ptr = TTY_PARITY;
			tty->flip.char_buf_ptr++;
			tty->flip.flag_buf_ptr++;
			tty->flip.count++;
		}
		events &= ~EVENT_PE;
	}
	if (events & EVENT_FE) {
		state->icount.frame += htonl(e->fe_count);
		if (info->read_status_mask & UART_LSR_FE) {
			++push;
			*tty->flip.char_buf_ptr = 0;
			*tty->flip.flag_buf_ptr = TTY_FRAME;
			tty->flip.char_buf_ptr++;
			tty->flip.flag_buf_ptr++;
			tty->flip.count++;
		}
		events &= ~EVENT_FE;
	}
	if (events & EVENT_BREAK) {
		state->icount.brk += htonl(e->brk_count);
		if (info->read_status_mask & UART_LSR_BI) {
			++push;
			*tty->flip.char_buf_ptr = 0;
			*tty->flip.flag_buf_ptr = TTY_BREAK;
			tty->flip.char_buf_ptr++;
			tty->flip.flag_buf_ptr++;
			tty->flip.count++;
		}
		events &= ~EVENT_BREAK;
	}
	/* push events into tty layer */
	spin_unlock_irqrestore(&map->fliplock, flags);
	if (push)
		tty_flip_buffer_push(tty);
	if (events & EVENT_CTS) {
		state->icount.cts++;
		events &= ~EVENT_CTS;
	}
	if (events & EVENT_DSR) {
		state->icount.dsr++;
		events &= ~EVENT_DSR;
	}
	if (events & EVENT_RING) {
		state->icount.rng++;
		events &= ~EVENT_RING;
	}
	if (events & EVENT_DCD) {
		state->icount.dcd++;
		if (info->flags & ASYNC_CHECK_CD) {
#if (defined(SERIAL_DEBUG_OPEN) || defined(SERIAL_DEBUG_INTR))
			printk("bhnS%d CD now %s...", info->line,
				(status & UART_MSR_DCD) ? "on" : "off");
#endif
			if (status & UART_MSR_DCD)
				wake_up_interruptible(&info->open_wait);
			else if (!((info->flags & ASYNC_CALLOUT_ACTIVE) &&
					(info->flags & ASYNC_CALLOUT_NOHUP))) {
#ifdef SERIAL_DEBUG_OPEN
				printk("doing serial hangup...");
#endif
				if (info->tty)
					tty_hangup(info->tty);
			}
#ifdef SERIAL_DEBUG_OPEN
			printk("\n");
#endif
		}

		events &= ~EVENT_DCD;
	}
	if (events & EVENT_RX_XON)
		events &= ~EVENT_RX_XON;
	if (events & EVENT_RX_XOFF)
		events &= ~EVENT_RX_XOFF;
	/* if anyone is looping waiting on this bit, get them going again */
	if (events & EVENT_TXFIN) {
#ifdef SERIAL_DEBUG_EVENTS
		printk("bhn: Setting UART_LSR_TEMT on %d\n", minor);
#endif
		uart_lsr[minor] |= UART_LSR_TEMT;
		events &= ~EVENT_TXFIN;
	}
	if (events)
		printk("bhn: %s: unhandled events %x\n", __FUNCTION__, events);
	/* Push events into tty subsystem */
	if (push) {
		save_flags(flags); cli();
		tty_flip_buffer_push(tty);
		restore_flags(flags);
	}
}

/*
 * bhn_wait_until_sent() --- wait until the transmitter is empty
 */
static void bhn_wait_until_sent(struct tty_struct *tty, int timeout)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	unsigned long orig_jiffies, char_time;
	int err;
	int minor = MINOR(tty->device);

	/* if not already empty */
	if (!(uart_lsr[minor] & UART_LSR_TEMT)) {
		info->flags |= ASYNC_QUERY_IN_PROGRESS;
		err = prot_query_tx_finished(minor);
		info->flags &= ~ASYNC_QUERY_IN_PROGRESS;
		if (err < 0) {
#ifdef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
			printk("%s: %d err=%d\n", __FUNCTION__, minor, err);
#endif
			return;
		}
	}
	else
		return;

	orig_jiffies = jiffies;
/*
 * With BH Net, we can't use the fifo size to determine the check interval.
 * Make it one character time.
 *
 */
#ifdef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
	printk("wait_until: info->timeout %d timeout %d HZ %d baud_base %d\n", info->timeout, timeout, HZ, info->state->baud_base );
#endif
	char_time = (info->timeout - HZ / 50) / (4096 + info->xmit_fifo_size);
	char_time = char_time / 5;
	if (char_time == 0)
		char_time = 1;
	if (timeout && timeout < char_time)
		char_time = timeout;
	/*
	 * If the transmitter hasn't cleared in twice the approximate
	 * amount of time to send the entire FIFO, it probably won't
	 * ever clear.  This assumes the UART isn't doing flow
	 * control, which is currently the case.  Hence, if it ever
	 * takes longer than info->timeout, this is probably due to a
	 * UART bug of some kind.  So, we clamp the timeout parameter at
	 * 2*info->timeout.
	 */
	if (!timeout || timeout > 2 * info->timeout)
		timeout = 2 * info->timeout;
#ifdef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
	printk("In bhn_wait_until_sent(%d) check=%lu...", timeout,
	       char_time);
	printk("jiff=%lu...\n", jiffies);
#endif
/* djw events are now in place, check to see UART_LSR_TEMT bit comes on */

	while (!(uart_lsr[minor] & UART_LSR_TEMT)) {
#ifdef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
		printk("wait: lsr=%d (jiff=%lu)...", uart_lsr[minor], jiffies);
#endif
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(char_time);
		if (signal_pending(current))
			break;
		if (timeout && time_after(jiffies, orig_jiffies + timeout))
			break;
	}
#ifdef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
	printk("lsr = %d (orig_jiff=%lu,jiff=%lu)...done\n",
		uart_lsr[minor], orig_jiffies, jiffies);
#endif
}

static int bhn_write(struct tty_struct *tty, int from_user,
					const unsigned char *buf, int count)
{
	struct async_struct *info = tty->driver_data;
	int minor = MINOR(tty->device);
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct bhn_connection *data = &state->data;
	int copied = 0;
#ifdef 	SERIAL_DEBUG_WRITE
	printk("%s bhnS%d - bytes %d user? %d\n", __FUNCTION__, minor, count,
								from_user);
#endif

#ifdef SERIAL_DEBUG_EVENTS
	printk("%s bhnS%d - resetting UART_LSR_TEMT\n", __FUNCTION__, minor);
#endif
	uart_lsr[minor] &= ~UART_LSR_TEMT;

	while (count) {
		int copy;

		copy = CIRC_SPACE_TO_END(data->write_head, data->write_tail,
								WRITE_BUF_SZ);
		if (!copy)
			break;

		if (count < copy)
			copy = count;

		if (from_user) {
			int ret;

			ret = copy_from_user(data->write_buf + data->write_head,
							buf + copied, copy);
			if (ret) {
				if (!copied) {
					copied = ret;
					goto err_out;
				} else
					goto copied_out;
			}
		} else
			memcpy(data->write_buf + data->write_head,
							buf + copied, copy);

		data->write_head = (data->write_head + copy) % WRITE_BUF_SZ;
		copied += copy;
		count -= copy;
	}

copied_out:
	if (info->state)
		info->state->icount.tx += copied;
	wake_up(&data->wq_write);

err_out:
	return copied;
}

static int bhn_ioctl(struct tty_struct *tty, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
	struct async_struct * info = (struct async_struct *)tty->driver_data;
	struct async_icount cprev, cnow;	/* kernel counter temps */
	struct serial_icounter_struct icount;
	unsigned long flags;
#if (LINUX_VERSION_CODE < 131394) /* Linux 2.1.66 */
	int retval, tmp;
#endif

	if ((cmd != TIOCGSERIAL) && (cmd != TIOCSSERIAL) &&
	    (cmd != TIOCSERCONFIG) && (cmd != TIOCSERGSTRUCT) &&
	    (cmd != TIOCMIWAIT) && (cmd != TIOCGICOUNT)) {
		if (tty->flags & (1 << TTY_IO_ERROR))
		    return -EIO;
	}

	switch (cmd) {
#if (LINUX_VERSION_CODE < 131394) /* Linux 2.1.66 */
		case TCSBRK:	/* SVID version: non-zero arg --> no break */
			retval = tty_check_change(tty);
			if (retval)
				return retval;
			tty_wait_until_sent(tty, 0);
			if (signal_pending(current))
				return -EINTR;
			if (!arg) {
				send_break(info, HZ/4);	/* 1/4 second */
				if (signal_pending(current))
					return -EINTR;
			}
			return 0;
		case TCSBRKP:	/* support for POSIX tcsendbreak() */
			retval = tty_check_change(tty);
			if (retval)
				return retval;
			tty_wait_until_sent(tty, 0);
			if (signal_pending(current))
				return -EINTR;
			send_break(info, arg ? arg*(HZ/10) : HZ/4);
			if (signal_pending(current))
				return -EINTR;
			return 0;
		case TIOCGSOFTCAR:
			tmp = C_CLOCAL(tty) ? 1 : 0;
			if (copy_to_user((void *)arg, &tmp, sizeof(int)))
				return -EFAULT;
			return 0;
		case TIOCSSOFTCAR:
			if (copy_from_user(&tmp, (void *)arg, sizeof(int)))
				return -EFAULT;

			tty->termios->c_cflag =
				((tty->termios->c_cflag & ~CLOCAL) |
				 (tmp ? CLOCAL : 0));
			return 0;
#endif
		case TIOCMGET:
			return get_modem_info(info, (unsigned int *) arg);
		case TIOCMBIS:
		case TIOCMBIC:
		case TIOCMSET:
			return set_modem_info(info, cmd, (unsigned int *) arg);
		case TIOCGSERIAL:
			return get_serial_info(info,
					       (struct serial_struct *) arg);
			return 0;
		case TIOCSSERIAL:
			return set_serial_info(info,
					       (struct serial_struct *) arg);
			return 0;
		case TIOCSERCONFIG:
			/* return do_autoconfig(info); */
			return 0;

		case TIOCSERGETLSR: /* Get line status register */
			return get_lsr_info(info, (unsigned int *) arg);

		case TIOCSERGSTRUCT:
			if (copy_to_user((struct async_struct *) arg,
					 info, sizeof(struct async_struct)))
				return -EFAULT;
			return 0;


#if 0
#ifdef CONFIG_SERIAL_MULTIPORT
		case TIOCSERGETMULTI:
			return get_multiport_struct(info,
				       (struct serial_multiport_struct *) arg);
		case TIOCSERSETMULTI:
			return set_multiport_struct(info,
				       (struct serial_multiport_struct *) arg);
#endif
#endif

		/*
		 * Wait for any of the 4 modem inputs (DCD,RI,DSR,CTS) to change
		 * - mask passed in arg for lines of interest
 		 *   (use |'ed TIOCM_RNG/DSR/CD/CTS for masking)
		 * Caller should use TIOCGICOUNT to see which one it was
		 */
		case TIOCMIWAIT:
			save_flags(flags); cli();
			/* note the counters on entry */
			cprev = info->state->icount;
			restore_flags(flags);
			/* Force modem status interrupts on */
			info->IER |= UART_IER_MSI;
			while (1) {
				interruptible_sleep_on(&info->delta_msr_wait);
				/* see if a signal did it */
				if (signal_pending(current))
					return -ERESTARTSYS;
				save_flags(flags); cli();
				cnow = info->state->icount; /* atomic copy */
				restore_flags(flags);
				if (cnow.rng == cprev.rng && cnow.dsr == cprev.dsr &&
				    cnow.dcd == cprev.dcd && cnow.cts == cprev.cts)
					return -EIO; /* no change => error */
				if ( ((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
				     ((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
				     ((arg & TIOCM_CD)  && (cnow.dcd != cprev.dcd)) ||
				     ((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) ) {
					return 0;
				}
				cprev = cnow;
			}
			/* NOTREACHED */

		/*
		 * Get counter of input serial line interrupts (DCD,RI,DSR,CTS)
		 * Return: write counters to the user passed counter struct
		 * NB: both 1->0 and 0->1 transitions are counted except for
		 *     RI where only 0->1 is counted.
		 */
		case TIOCGICOUNT:
			save_flags(flags); cli();
			cnow = info->state->icount;
			restore_flags(flags);
			icount.cts = cnow.cts;
			icount.dsr = cnow.dsr;
			icount.rng = cnow.rng;
			icount.dcd = cnow.dcd;
			icount.rx = cnow.rx;
			icount.tx = cnow.tx;
			icount.frame = cnow.frame;
			icount.overrun = cnow.overrun;
			icount.parity = cnow.parity;
			icount.brk = cnow.brk;
			icount.buf_overrun = cnow.buf_overrun;

			if (copy_to_user((void *)arg, &icount, sizeof(icount)))
				return -EFAULT;
			return 0;
		case TIOCSERGWILD:
		case TIOCSERSWILD:
			/* "setserial -W" is called in Debian boot */
			printk ("TIOCSER?WILD ioctl obsolete, ignored.\n");
			return 0;

		default:
			return -ENOIOCTLCMD;
		}
	return 0;
}

static void bhn_set_termios(struct tty_struct *tty, struct termios *old_term)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	unsigned int cflag = tty->termios->c_cflag;
	int minor = MINOR(tty->device);

	if ((cflag == old_term->c_cflag) &&
				(RELEVANT_IFLAG(tty->termios->c_iflag) ==
					RELEVANT_IFLAG(old_term->c_iflag)))
		return;

	change_speed(info, old_term);

	/* Handle transition to B0 status */
	if ((old_term->c_cflag & CBAUD) && !(cflag & CBAUD)) {
		info->MCR &= ~(UART_MCR_DTR | UART_MCR_RTS);
		prot_set_rts_dtr(minor, 0, 0);
	}

	/* Handle transition away from B0 status */
	if (!(old_term->c_cflag & CBAUD) && (cflag & CBAUD)) {
		info->MCR |= UART_MCR_DTR;
		if (!(tty->termios->c_cflag & CRTSCTS) ||
					!test_bit(TTY_THROTTLED, &tty->flags))
			info->MCR |= UART_MCR_RTS;
		prot_set_rts_dtr(minor, info->MCR & UART_MCR_RTS, 1);
	}

	/* Handle turning off CRTSCTS */
	if ((old_term->c_cflag & CRTSCTS) &&
					!(tty->termios->c_cflag & CRTSCTS)) {
		tty->hw_stopped = 0;
		bhn_start(tty);
	}
}

static void bhn_flush_chars(struct tty_struct *tty)
{
	/* nop */
}

static int bhn_write_room(struct tty_struct *tty)
{
	int minor = MINOR(tty->device);
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_connection *data = &map->state->data;

	return CIRC_SPACE(data->write_head, data->write_tail, WRITE_BUF_SZ);
}

static int bhn_chars_in_buffer(struct tty_struct *tty)
{
	int minor = MINOR(tty->device);
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_connection *data = &map->state->data;

	return CIRC_CNT(data->write_head, data->write_tail, WRITE_BUF_SZ);
}


static void bhn_throttle(struct tty_struct *tty)
{
	struct async_struct *info = tty->driver_data;
	int minor = MINOR(tty->device);

#ifdef SERIAL_DEBUG_THROTTLE
	char buf[64];

	printk("throttle %s: %d\n", tty_name(tty, buf),
					tty->ldisc.chars_in_buffer(tty));
#endif

	/* no flow control enabled */
	if (!I_IXOFF(tty) && !(tty->termios->c_cflag & CRTSCTS))
		return;

	if (I_IXOFF(tty))
		info->x_char = STOP_CHAR(tty);

	if (tty->termios->c_cflag & CRTSCTS) {
		info->MCR &= ~UART_MCR_RTS;
		prot_set_rts(minor, 0);
	}

	bhn_table[minor].flags |= ASYNC_THROTTLE;
}

static void bhn_unthrottle(struct tty_struct *tty)
{
	struct async_struct *info = tty->driver_data;
	int minor = MINOR(tty->device);

#ifdef SERIAL_DEBUG_THROTTLE
	char buf[64];

	printk("unthrottle %s: %d\n", tty_name(tty, buf),
					tty->ldisc.chars_in_buffer(tty));
#endif

	bhn_table[minor].flags &= ~ASYNC_THROTTLE;
	wake_up_interruptible(&bhn_table[minor].state->wq_throttle);

	/* no flow control enabled */
	if (!I_IXOFF(tty) && !(tty->termios->c_cflag & CRTSCTS))
		return;

	if (I_IXOFF(tty))
		info->x_char = 0;

	if (tty->termios->c_cflag & CRTSCTS) {
		info->MCR |= UART_MCR_RTS;
		prot_set_rts(minor, 1);
	}
}

static void bhn_stop(struct tty_struct *tty)
{
#ifdef SERIAL_DEBUG_START
	printk("%s - %d\n", __FUNCTION__, MINOR(tty->device));
#endif
}

static void bhn_start(struct tty_struct *tty)
{
#ifdef SERIAL_DEBUG_START
	printk("%s - port %d\n", __FUNCTION__, MINOR(tty->device));
#endif
}

static void bhn_hangup(struct tty_struct *tty)
{
	struct async_struct * info = tty->driver_data;
	struct serial_state *state = info->state;
	int minor = MINOR(tty->device);
	struct bhn_map *map = &bhn_table[minor];

	bhn_flush_buffer(tty);
	if (info->flags & ASYNC_CLOSING)
		return;
	shutdown(info);
	down(&map->auto_reconn_lock);
	if (map->auto_reconn_pid) {
		up(&map->auto_reconn_lock);
		kill_proc(map->auto_reconn_pid, SIGKILL, 1);
		wait_for_completion(&map->auto_reconn_comp);
	} else
		up(&map->auto_reconn_lock);
	bhn_network_down(minor);
	info->event = 0;
	state->count = 0;
	info->flags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CALLOUT_ACTIVE);
	info->tty = 0;
	wake_up_interruptible(&info->open_wait);
}

#if (LINUX_VERSION_CODE >= 0x020142)
static void bhn_break(struct tty_struct *tty, int onoff)
{
    int minor = MINOR(tty->device);

    prot_start_stop_break(minor, onoff == -1 ? 1 : 0);

	return;
}
#endif

static void bhn_flush_buffer(struct tty_struct *tty)
{
	int minor = MINOR(tty->device);
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_state *state = map->state;
	struct bhn_connection *data = &state->data;

	down(&map->auto_reconn_lock);
	map->flags |= MAP_PURGING;
	up(&map->auto_reconn_lock);
	data->write_head = data->write_tail = 0;
	/* prot_purge turns off the PURGING flag, so we don't have to */
	prot_purge(minor, PURGE_WRITE);

	wake_up_interruptible(&tty->write_wait);
#ifdef SERIAL_HAVE_POLL_WAIT
	wake_up_interruptible(&tty->poll_wait);
#endif
	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
							tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup)(tty);
}

static void bhn_send_xchar(struct tty_struct *tty, char c)
{
    int minor = MINOR(tty->device);
    char xchar = 0;
	struct async_struct *info = (struct async_struct *)tty->driver_data;


    info->x_char = c;

    if(c == tty->termios->c_cc[VSTART])
        xchar = XCHAR_XON;
    else if(c == tty->termios->c_cc[VSTOP])
        xchar = XCHAR_XOFF;
    else
        return;

    prot_send_xchar(minor, xchar);

	return;
}


#ifdef CONFIG_PROC_FS
/*
 * /proc fs routines....
 */

static inline int line_info(int minor, char *buf, struct serial_state *state)
{
	struct async_struct *info = state->info;
	char stat_buf[30];
	int ret;

	struct bhn_map *map = &bhn_table[minor];
	struct bhn_unit *unit = map->unit;
	struct bhn_info *binfo = &unit->info;

	ret = sprintf(buf, "%3d: IP %d.%d.%d.%d TCP %d port %d", minor,
					binfo->bhn_ip[0], binfo->bhn_ip[1],
					binfo->bhn_ip[2], binfo->bhn_ip[3],
					CMDPORT(map), map->unit_port);

	/*
	 * Figure out the current RS-232 lines
	 */

	stat_buf[0] = 0;
	stat_buf[1] = 0;
	if (info && info->MCR & UART_MCR_RTS)
		strcat(stat_buf, "|RTS");
	if (uart_msr[minor] & UART_MSR_CTS)
		strcat(stat_buf, "|CTS");
	if (info && info->MCR & UART_MCR_DTR)
		strcat(stat_buf, "|DTR");
	if (uart_msr[minor] & UART_MSR_DSR)
		strcat(stat_buf, "|DSR");
	if (uart_msr[minor] & UART_MSR_DCD)
		strcat(stat_buf, "|CD");
	if (uart_msr[minor] & UART_MSR_RI)
		strcat(stat_buf, "|RI");
	if (info && info->quot)
		ret += sprintf(buf + ret, " baud:%d",
						state->baud_base / info->quot);

	ret += sprintf(buf + ret, " tx:%d rx:%d",
		       state->icount.tx, state->icount.rx);

	if (state->icount.frame)
		ret += sprintf(buf + ret, " fe:%d", state->icount.frame);

	if (state->icount.parity)
		ret += sprintf(buf + ret, " pe:%d", state->icount.parity);

	if (state->icount.brk)
		ret += sprintf(buf + ret, " brk:%d", state->icount.brk);

	if (state->icount.overrun)
		ret += sprintf(buf + ret, " oe:%d", state->icount.overrun);

	/*
	 * Last thing is the RS-232 status lines
	 */
	ret += sprintf(buf + ret, " %s\n", stat_buf + 1);
	return ret;
}

static int bhn_read_proc(char *page, char **start, off_t off, int count,
			 int *eof, void *data)
{
	int i, len = 0;
	off_t begin = 0;

	len += sprintf(page, "%s driver version %s (%s)\n", bhn_drvname,
		       bhn_version, bhn_revdate);

	for (i = 0; i < NPORTS && len < 4000; i++) {
		/* only display information for ports that are mapped */
		if (bhn_table[i].unit) {
			len += line_info(i, page + len, &rs_table[i]);
			if (len + begin > off + count)
				goto done;
			if (len + begin < off) {
				begin += len;
				len = 0;
			}
		}
	}
done:
	*eof = 1;
	if (off >= len + begin)
		return 0;
	*start = page + (off - begin);
	return ((count < begin + len - off) ? count : begin + len - off);
}

static int bhn_write_proc(struct file *file, const char *buffer,
			  unsigned long count, void *data)
{
	return 0;
}

#define PROC_STATE_START	0
#define PROC_STATE_RUN		1
#define PROC_STATE_EOF		2

#define PRINT_CONFIG_HELP(p)	{										\
		p += sprintf(p, "# A local host device node is mapped to a remote BlueHeat/Net serial\n");	\
		p += sprintf(p, "# port with a mapping line. The fields on a mapping line are:\n");		\
		p += sprintf(p, "#\n");										\
		p += sprintf(p, "# minor IP TCP-base-port unit-serial-port cmd-timeout auto-reconnect\n");	\
		p += sprintf(p, "#\n");										\
		p += sprintf(p, "# minor\t\tThe local minor number to use. This is also the XX of\n");		\
		p += sprintf(p, "#\t\t/dev/bhnSXX.\n");								\
		p += sprintf(p, "# IP\t\tThe IP address of the BlueHeat/Net to be mapped.\n");			\
		p += sprintf(p, "# TCP-base-port\tThe TCP base port of the BlueHeat/Net to be mapped.\n");	\
		p += sprintf(p, "#\t\tSee your BlueHeat/Net documentation for more info.\n");			\
		p += sprintf(p, "# unit-serial-port\tThe 1-based number of the serial port to mapped,\n");	\
		p += sprintf(p, "#\t\t\ton the BlueHeat/Net.\n");							\
		p += sprintf(p, "# cmd-timeout\tHow long to wait for responses from the mapped port\n");	\
		p += sprintf(p, "#\t\tbefore deciding the network connection is dead. This is in\n");		\
		p += sprintf(p, "#\t\ttenths of a second. 0 waits forever.\n");					\
		p += sprintf(p, "# auto-reconnect\t0 - disable, 1 - enable. When enabled will attempt\n");	\
		p += sprintf(p, "#\t\t\tto automatically reconnect with the port after a\n");		\
		p += sprintf(p, "#\t\t\tnetwork disruption.\n");								\
		p += sprintf(p, "#\n");										\
		p += sprintf(p, "# Examples:\n");								\
		p += sprintf(p, "#\n");										\
		p += sprintf(p, "# 4 10.0.0.6 49152 3 300 0\n");						\
		p += sprintf(p, "# /dev/bhnS4 is mapped to port 3 on BlueHeat/Net 10.0.0.6, TCP port\n");	\
		p += sprintf(p, "# 49152. Command timeout is 30 seconds, autoreconnect is off.\n");		\
		p += sprintf(p, "#\n");										\
		p += sprintf(p, "# 0 192.168.7.1 30000 8 20 1\n");						\
		p += sprintf(p, "# /dev/bhnS0 is mapped to port 8 on BlueHeat/Net 192.168.7.1, TCP port\n");	\
		p += sprintf(p, "# 30000. Command timeout is 2 seconds, autoreconnect is on.\n");		\
	}

/*
 * A note about proc_read routines and their paramters:
 *
 * start and eof are output parameters, all the rest are inputs.
 *
 * page:
 *
 * The address of the start of the page to whence output should be
 * written. It will remain the same until either eof is set or a
 * 0 length output is returned. There is only one page, one must not
 * write past the end.
 *
 * start:
 *
 * The function of this parameter is to allow certain file-position
 * manipulations from the routine. Since we're not interested in that
 * it should always be set to page.
 *
 * off:
 *
 * The actual file-position. If the routine is called with a new page,
 * this is will remain incremented and attempting to output to page + off
 * will result in corruption. When operating in fresh-page mode (see eof)
 * off can safely be ignored.
 *
 * count:
 *
 * How many bytes of space there are left in page. The proc fs tries to
 * account for bad read routines by lying about the count; while there
 * is greater than page - 1024 bytes of space left it will report only
 * page - 1024. This puzzled me for some time, until I looked at the
 * code of the calling routine. Exceeding this value means that you've
 * written past the end of the page, and corrupted something.
 *
 * eof:
 *
 * proc_read routines are called from a loop inside the kernel. The
 * purpose is to fill the page with output before returning it to user
 * space. To this end, setting eof does not actually signal eof to user
 * space. All it does is terminate the in-kernel loop. Thus eof is really
 * return-current-output-to-user-space-now-please. Always setting eof
 * means that the output you've just generated will be returned right away
 * to user space. If there's another call to your routine, it corresponds
 * to another read() from user space. Also, it will have a fresh page to
 * write output to. Also, returning a 0 length output will always signal
 * actual eof, since it's guaranteed that the 0 length output will be on
 * a fresh page.
 *
 * data:
 *
 * A private data pointer; not in use by our driver at this time.
 *
 * return value:
 *
 * Return the amount of data generated. A positive count will cause the
 * data, to be returned to user space. A negative count will corrupt the
 * caller, so don't do that. A 0 count will signal that the routine is
 * done outputing for now, and return all the accumlatated output so far
 * to user space. If you've written data to the page in previous calls the
 * user space read() will get some data back, and in some cases will
 * initiate a new read(). This will then initiate a new call to the
 * routine. The way to signify eof is to return 0 from a page that has not
 * had any output written to it. The easiest way to do this is to use
 * eof to run in "fresh page mode". See eof.
 */
static int bhn_proc_autod_read(char *page, char **start, off_t off, int count,
							int *eof, void *data)
{
	static int state = PROC_STATE_START;
	static char *saved_page;
	static char *p;
	static int saved_count = 0;
	static int minor;
	struct bhn_info info;
	int port;
	int len;
	int err;

	if (!off && state != PROC_STATE_START) {
		if (saved_count) {
			saved_count = 0;
			free_page((unsigned long)saved_page);
		}
		if (state == PROC_STATE_RUN)
			prot_stop_autod();
		state = PROC_STATE_START;
	} else if (saved_count)
		goto print_out;

	if (state == PROC_STATE_START) {
		saved_page = (char *)get_zeroed_page(GFP_KERNEL);
		if (!saved_page)
			return -ENOMEM;

		minor = 0;

		p = saved_page;
		p += sprintf(p, "#\n");
		p += sprintf(p, "# Any of the following parameters may be set by writing them. eg\n");
		p += sprintf(p, "# 'cat \"%s=0\" > /proc/%s/%s'\n", AUTOD_TIMEOUT_NM, PROC_BHNROOT_NM, PROC_AUTOD_NM);
		p += sprintf(p, "#\n");
		p += sprintf(p, "# The current values are:\n");
		p += sprintf(p, "#\n");
		p += sprintf(p, "# %s=%d.%d.%d.%d\n", AUTOD_NETWORK_NM, autod_netaddr[0], autod_netaddr[1], autod_netaddr[2], autod_netaddr[3]);
		p += sprintf(p, "# %s=%d\n", AUTOD_PORT_NM, autod_port);
		p += sprintf(p, "# %s=%d\n", AUTOD_POLL_NM, autod_poll);
		p += sprintf(p, "# %s=%d\n", AUTOD_TIMEOUT_NM, autod_timeout);
		p += sprintf(p, "#\n");
		p += sprintf(p, "# %s\tThe network address on which to send the autodiscover broadcast.\n", AUTOD_NETWORK_NM);
		p += sprintf(p, "# %s\t\tThe TCP port to send the autodiscover broadcast to. This should\n", AUTOD_PORT_NM);
		p += sprintf(p, "# \t\tmatch the TCP base port of the BlueHeat/Nets to be found.\n");
		p += sprintf(p, "# %s\t\tHow often to poll for a response from a BlueHeat/Net.\n", AUTOD_POLL_NM);
		p += sprintf(p, "# %s\tHow long to run the autodiscover for.\n", AUTOD_TIMEOUT_NM);
		p += sprintf(p, "#\n");
		p += sprintf(p, "# %s and %s are in tenths of a second. %s must be between\n", AUTOD_POLL_NM, AUTOD_TIMEOUT_NM, AUTOD_POLL_NM);
		p += sprintf(p, "# 1 and %s inclusive. %s must be equal or greater than %s.\n", AUTOD_TIMEOUT_NM, AUTOD_TIMEOUT_NM, AUTOD_POLL_NM);
		p += sprintf(p, "# %s may also be 0 (wait forever) or -1 (print header only)\n", AUTOD_TIMEOUT_NM);
		p += sprintf(p, "#\n");
		PRINT_CONFIG_HELP(p);
		p += sprintf(p, "#\n");
		if (autod_timeout == -1)
			state = PROC_STATE_EOF;
		else {
			p += sprintf(p, "# Autodiscovering Blue Heat/Nets...\n");
			p += sprintf(p, "#\n");
			err = prot_start_autod(autod_port,
				*((unsigned int *)autod_netaddr), NULL, NULL);
			if (err < 0) {
				p += sprintf(p, "# Error starting autodiscovery: %s (%d)\n", (elist[-err] ? elist[-err] : "<no text>"), err);
				state = PROC_STATE_EOF;
			} else
				state = PROC_STATE_RUN;
		}

		saved_count = p - saved_page;
		p = saved_page;

		goto print_out;
	}

	if (state == PROC_STATE_EOF) {
		state = PROC_STATE_START;
		len = 0;
		goto out;
	}

	if (minor == 256)
		goto reset;

	err = prot_get_autod(&info, (autod_poll * HZ) / 10,
						(autod_timeout * HZ) / 10);
	if (err < 0)
		goto reset;

	saved_page = (char *)get_zeroed_page(GFP_KERNEL);
	if (!saved_page)
		return -ENOMEM;

	p = saved_page;
	p += sprintf(p, "\n# Blue Heat/Net (%d ports) %d.%d.%d.%d\n",
				info.num_ports, info.bhn_ip[0], info.bhn_ip[1],
				info.bhn_ip[2], info.bhn_ip[3]);
	p += sprintf(p, "# Serial #%d\tMAC %02x:%02x:%02x:%02x:%02x:%02x\n",
			ntohl(info.bhn_ser_num),
			info.bhn_mac[0], info.bhn_mac[1], info.bhn_mac[2],
			info.bhn_mac[3], info.bhn_mac[4], info.bhn_mac[5]);
	p += sprintf(p, "# Hostname %s.%s\n", info.bhn_hostname,
							info.bhn_domainname);

	for (port = 1; (port <= info.num_ports) && (minor < 256);
								port++, minor++)
		p += sprintf(p, AUTOD_FORMAT, minor, info.bhn_ip[0],
				info.bhn_ip[1], info.bhn_ip[2], info.bhn_ip[3],
				autod_port, port, DEFAULT_CMD_TIMEOUT,
				DEFAULT_AUTO_RECONNECT);

	saved_count = p - saved_page;
	p = saved_page;

print_out:
	len = MIN(saved_count, count);
	memcpy(page, p, len);

	saved_count -= len;
	if (!saved_count)
		free_page((unsigned long)saved_page);
	else
		p += len;

out:
	*eof = 1;
	*start = page;
	return len;

reset:
	prot_stop_autod();
	state = PROC_STATE_START;
	len = 0;
	goto out;
}

static int bhn_proc_autod_write(struct file *file, const char *buffer,
					unsigned long count, void *data)
{
	char buf[128];
	char *p;
	char *key;
	char *arg;
	int n1, n2, n3, n4;

	if (count > sizeof(buf) - 1)
		count = sizeof(buf) - 1;

	if (copy_from_user(buf, buffer, count))
		return -EFAULT;

	buf[count] = '\0';

	p = buf;
	while (*p) {
		SKIP_WSPACE(p);
		ASSIGN_OR_EOB(p, key);
		SKIP_TOKEN(p);
		SKIP_WSPACE(p);
		EQUALS(p);
		SKIP_WSPACE(p);
		ASSIGN(p, arg);
		SKIP_TOKEN(p);

		if (!strncmp(AUTOD_PORT_NM, key, AUTOD_PORT_SZ)) {
			sscanf(arg, "%d", &n1);
			if ((n1 >= 0) && (n1 <= 65535))
				autod_port = n1;
		} else if (!strncmp(AUTOD_POLL_NM, key, AUTOD_POLL_SZ)) {
			sscanf(arg, "%d", &n1);
			if ((n1 >= 1) && (n1 <= autod_timeout))
				autod_poll = n1;
		} else if (!strncmp(AUTOD_TIMEOUT_NM, key, AUTOD_TIMEOUT_SZ)) {
			sscanf(arg, "%d", &n1);
			if ((n1 >= autod_poll) || (!n1) || (n1 == -1))
				autod_timeout = n1;
		} else if (!strncmp(AUTOD_NETWORK_NM, key, AUTOD_NETWORK_SZ)) {
			/*
			 * NOTE: This must be done in two stages because there's
			 * no way to tell sscanf() to reduce the %d to the size
			 * of a char before assigning the value.
			 */
			sscanf(arg, "%d.%d.%d.%d", &n1, &n2, &n3, &n4);
			autod_netaddr[0] = (unsigned char)n1;
			autod_netaddr[1] = (unsigned char)n2;
			autod_netaddr[2] = (unsigned char)n3;
			autod_netaddr[3] = (unsigned char)n4;
		} else
			return -EINVAL;
	}

	return count;
}

static int bhn_proc_config_read(char *page, char **start, off_t off, int count,
							int *eof, void *data)
{
	static int state = PROC_STATE_START;
	static char *saved_page;
	static char *p;
	static int saved_count = 0;
	static int uidx;
	struct bhn_unit *unit;
	struct bhn_map *map;
	struct bhn_info *info;
	int cmd_timeout, auto_reconnect;
	int port;
	int len;

	if (!off && state != PROC_STATE_START) {
		if (saved_count) {
			saved_count = 0;
			free_page((unsigned long)saved_page);
		}
		state = PROC_STATE_START;
	} else if (saved_count)
		goto print_out;

	if (state == PROC_STATE_START) {
		saved_page = (char *)get_zeroed_page(GFP_KERNEL);
		if (!saved_page)
			return -ENOMEM;

		uidx = 0;

		p = saved_page;
		p += sprintf(p, "#\n");
		PRINT_CONFIG_HELP(p);
		p += sprintf(p, "#\n");
		p += sprintf(p, "# Blue Heat/Net Linux Host Driver configuration\n");
		p += sprintf(p, "#\n");
		state = PROC_STATE_RUN;

		saved_count = p - saved_page;
		p = saved_page;

		goto print_out;
	}

	unit = &bhn_unit_table[uidx];
	if (!unit->base_baud) {
		state = PROC_STATE_START;
		len = 0;
		goto out;
	}

	saved_page = (char *)get_zeroed_page(GFP_KERNEL);
	if (!saved_page)
		return -ENOMEM;

	p = saved_page;
	p += sprintf(p, "\n# Blue Heat/Net #%d\n", uidx);
	uidx++;

	info = &unit->info;
	for (port = 0; port < MAX_PORTS; port++) {
		map = unit->ports[port];
		if (!map)
			continue;
		if (map->cmd_timeout == MAX_SCHEDULE_TIMEOUT)
			cmd_timeout = 0;
		else
			cmd_timeout = map->cmd_timeout / (HZ / 10);
		auto_reconnect = (map->flags & MAP_AUTO_RECONNECT) ? 1 : 0;
		p += sprintf(p, AUTOD_FORMAT, MAP_TO_MINOR(map),
					info->bhn_ip[0], info->bhn_ip[1],
					info->bhn_ip[2], info->bhn_ip[3],
					unit->base_port, port + 1,
					cmd_timeout, auto_reconnect);
	}

	saved_count = p - saved_page;
	p = saved_page;

print_out:
	len = MIN(saved_count, count);
	memcpy(page, p, len);

	saved_count -= len;
	if (!saved_count)
		free_page((unsigned long)saved_page);
	else
		p += len;

out:
	*eof = 1;
	*start = page;
	return len;
}

static void unmap_port(int minor) {
	struct bhn_map *map = &bhn_table[minor];
	struct bhn_unit *unit;
	int i;
	int j;

	if (!map->unit)
		return;

	unit = map->unit;
	map->unit = NULL;
	unit->ports[map->unit_port - 1] = NULL;
	for (i = 0; i < MAX_PORTS && !unit->ports[i]; i++);
	if (i == MAX_PORTS) {
		for (i = (unit - bhn_unit_table) + 1; i < NPORTS &&
					bhn_unit_table[i].base_baud; i++) {
			memmove(&bhn_unit_table[i - 1], &bhn_unit_table[i],
						sizeof(bhn_unit_table[0]));
			for (j = 0; j < MAX_PORTS; j++) {
				if (bhn_unit_table[i - 1].ports[j])
					bhn_unit_table[i - 1].ports[j]->unit =
							&bhn_unit_table[i - 1];
			}
		}
		memset(&bhn_unit_table[i - 1], 0, sizeof(bhn_unit_table[0]));
	}
}

static char *nextline(char **p, unsigned long *left) {
	char *start = *p;
	unsigned long count = *left;

	SKIPN_LINE(*p, *left);
	if (!*left) {
		*p = start;
		*left = count;
		return NULL;
	}

	**p = '\0';
        (*p)++;
	(*left)--;
	return start;
}

static int bhn_proc_config_write(struct file *file, const char *buffer,
					unsigned long count, void *data)
{
	static char *part_line = NULL;
	static unsigned long part_count = 0;
	char *buf;
	char *p;
	char *line, *oline;
	int lnum;
	unsigned long left;
	int minor, base_port, unit_port, cmd_timeout, auto_reconnect;
	int n1, n2, n3, n4;
	struct bhn_unit *unit;
	int i;

	if (!count)
		return count;

	buf = (char *)kmalloc(count + part_count, GFP_KERNEL);
	if (!buf)
		return -ENOMEM;

	if (part_count)
		memcpy(buf, part_line, part_count);
	if (copy_from_user(buf + part_count, buffer, count))
		return -EFAULT;

	p = buf;
	left = count + part_count;
	lnum = 0;
	for (line = nextline(&p, &left); line; line = nextline(&p, &left)) {
		lnum++;
		oline = line;
		SKIP_WSPACE(line);
		SKIP_COMMENT(line);
		SKIP_EMPTY(line);
		if (!memcmp("clear-mappings", line, 14)) {
			for (i = 0; i < NPORTS; i++) {
				if (bhn_table[i].state)
					printk("%s: bhnS%d is busy, cannot "
						"unmap\n", bhn_drvname, i);
				else
					unmap_port(i);
			}
			continue;
		}
		cmd_timeout = DEFAULT_CMD_TIMEOUT;
		auto_reconnect = DEFAULT_AUTO_RECONNECT;
		sscanf(line, AUTOD_FORMAT, &minor, &n1, &n2, &n3, &n4,
					&base_port, &unit_port, &cmd_timeout,
					&auto_reconnect);
		if ((minor < 0) || (minor > 255)) {
			if ((lnum == 1) && (part_count))
				goto part_retry;
			printk("%s: Bad minor '%d'\n", bhn_drvname, minor);
			goto loop_err;
		}
		if ((n1 < 0) || (n1 > 255) || (n2 < 0) || (n2 > 255) ||
		    (n3 < 0) || (n3 > 255) || (n4 < 0) || (n4 > 255)) {
			if ((lnum == 1) && (part_count))
				goto part_retry;
			printk("%s: Bad IP '%d.%d.%d.%d'\n", bhn_drvname, n1,
								n2, n3, n4);
			goto loop_err;
		}
		if (bhn_table[minor].unit && (n1 || n2 || n3 || n4)) {
			if ((lnum == 1) && (part_count))
				goto part_retry;
			printk("%s: Minor '%d' already mapped\n", bhn_drvname,
									minor);
			goto loop_err;
		} else if (!bhn_table[minor].unit && !(n1 || n2 || n3 || n4)) {
			if ((lnum == 1) && (part_count))
				goto part_retry;
			printk("%s: Minor '%d' already unmapped\n", bhn_drvname,
									minor);
			goto loop_err;
		}
		if ((base_port < 0) || (base_port > 65535)) {
			if ((lnum == 1) && (part_count))
				goto part_retry;
			printk("%s: Bad base_port '%d'\n", bhn_drvname,
								base_port);
			goto loop_err;
		}
		if ((unit_port < 1) || (unit_port > MAX_PORTS)) {
			if ((lnum == 1) && (part_count))
				goto part_retry;
			printk("%s: Bad unit_port '%d'\n", bhn_drvname,
								unit_port);
			goto loop_err;
		}
		if (cmd_timeout < 0) {
			if ((lnum == 1) && (part_count))
				goto part_retry;
			printk("%s: Bad cmd_timeout '%d'\n", bhn_drvname,
								cmd_timeout);
			goto loop_err;
		}
		if (auto_reconnect < 0) {
			if ((lnum == 1) && (part_count))
				goto part_retry;
			printk("%s: Bad auto-reconnect '%d'\n", bhn_drvname,
								auto_reconnect);
			goto loop_err;
		}

		if (!n1 && !n2 && !n3 && !n4)
			unit = bhn_table[minor].unit;
		else {
			for (unit = bhn_unit_table, i = 0;
						unit->base_baud && i < NPORTS;
						unit++, i++)
				if (*(unsigned int *)unit->info.bhn_ip ==
							htonl((unsigned int)
							INADDR(n1,n2,n3,n4)))
					break;
			if (i == NPORTS) {
				if ((lnum == 1) && (part_count))
					goto part_retry;
				printk("%s: Unit table full\n", bhn_drvname);
				goto loop_err;
			}
		}
		if (unit->base_baud && (unit->base_port != base_port)) {
			if ((lnum == 1) && (part_count))
				goto part_retry;
			printk("%s: Unit has conflicting TCP base port '%d'\n",
				bhn_drvname, unit->base_port);
			goto loop_err;
		}
		if (n1 || n2 || n3 || n4) {
			if (unit->ports[unit_port - 1]) {
				if ((lnum == 1) && (part_count))
					goto part_retry;
				printk(
				"%s: Unit serial port '%d' already mapped\n",
							bhn_drvname, unit_port);
				goto loop_err;
			}
			if (!unit->base_baud) {
				*((int *)unit->info.bhn_ip) =
						htonl(INADDR(n1,n2,n3,n4));
				unit->base_port = base_port;
				/* FIXME: This isn't always correct */
				unit->base_baud = 460800;
				unit->addr.sin_family = PF_INET;
				unit->addr.sin_port = htons(unit->base_port);
				unit->addr.sin_addr.s_addr =
						 *((int *)unit->info.bhn_ip);
			}
			unit->ports[unit_port - 1] = &bhn_table[minor];
			bhn_table[minor].unit = unit;
			bhn_table[minor].unit_port = unit_port;
			if (!cmd_timeout)
				bhn_table[minor].cmd_timeout =
							MAX_SCHEDULE_TIMEOUT;
			else
				bhn_table[minor].cmd_timeout =
							cmd_timeout * (HZ / 10);
			if (auto_reconnect)
				bhn_table[minor].flags |= MAP_AUTO_RECONNECT;
			else
				bhn_table[minor].flags &= ~MAP_AUTO_RECONNECT;
		} else {
			if (bhn_table[minor].state) {
				if ((lnum == 1) && (part_count))
					goto part_retry;
				printk("%s: Port is busy, cannot unmap\n",
								bhn_drvname);
				goto loop_err;
			}
			unmap_port(minor);
		}
		continue;

part_retry:
		*(p - 1) = '\n';
		p = buf + part_count;
		left = count;
		lnum = 0;
		part_count = 0;
		kfree(part_line);
		continue;

loop_err:
		printk("%s: Error writing to /proc/%s/%s line %d '%s' \n",
				bhn_drvname, PROC_BHNROOT_NM, PROC_CONFIG_NM,
				lnum, oline);
	}

	if (part_count) {
		part_count = 0;
		kfree(part_line);
	}

	if (left) {
		if (!lnum)
			part_line = buf;
		else {
			part_line = (char *)kmalloc(left, GFP_KERNEL);
			if (!part_line) {
				part_line = buf;
				memmove(part_line, p, left);
			} else {
				memcpy(part_line, p, left);
				kfree(buf);
			}
		}
		part_count = left;
	} else
		kfree(buf);

	return count;
}
#endif

int receive_chars(int line, unsigned char *buf, int count)
{
	struct serial_state *state;
	struct async_struct *info;
	struct tty_struct *tty;
	struct bhn_map *map;
	struct	async_icount *icount;
	int put_count = 0;
	struct tty_ldisc *ldisc;
	int tty_room;
	unsigned long flags;

	state = rs_table + line;
	if (!state->info) {
		/* this port is probably not even open */
		printk("bhn: %s: No state on line %d\n", __FUNCTION__, line);
		goto out;
	}

	info = state->info;
	tty = info->tty;
	if (tty->flip.count)
		/* a previously push()ed buffer hasn't been handled yet */
		goto out;

	ldisc = tty_ldisc_ref(tty);
	if (!ldisc)
		goto out;
	if (!ldisc->receive_room) {
		tty_ldisc_deref(ldisc);
		goto out;
	}

	tty_room = ldisc->receive_room(tty);
	tty_ldisc_deref(ldisc);
	if (!tty_room)
		goto out;

	icount = &info->state->icount;

	put_count = TTY_FLIPBUF_SIZE;
	if (put_count > count)
		put_count = count;
	if (put_count > tty_room)
		put_count = tty_room;

	map = bhn_table + line;
	spin_lock_irqsave(&map->fliplock, flags);
	memcpy(tty->flip.char_buf_ptr, buf, put_count);
	tty->flip.char_buf_ptr += put_count;
	memset(tty->flip.flag_buf_ptr, 0, put_count);
	tty->flip.flag_buf_ptr += put_count;
	tty->flip.count += put_count;
	icount->rx += put_count;
	count -= put_count;

#ifdef SERIAL_DEBUG_READ
	printk("bhn: pushing %d latency %d interrupt %d ptr %ld\n", tty->flip.count, tty->low_latency, in_interrupt(), ((unsigned long)tty->flip.char_buf_ptr) -  ((unsigned long)&tty->flip.char_buf[0]));
#endif

	spin_unlock_irqrestore(&map->fliplock, flags);
	tty_flip_buffer_push(tty);

out:
	return count;
}
